
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. 

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
     curl [-s] <contents file>

   Date: May 1st, 1996

   Takes a 'contents' file, and generates a series of *linked* HTML
   files in the configuration given in the contents, using
   the unlinked HTML files. A KEYSHTML file is
   built for keywords, a FIGSHTML file for figures, and a
   TBLSHTML for tables, and a NEWHTML file for new pages.
   SRCHHTML is a forms interface to a search engine.

   curl will also work with HTML files previously generated by curl.

   The configuration of the HTML files can be altered by
   changing the 'contents' file and reexecuting curl.

   curl will also reprocess HTML files modified since
   contents was altered. 

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

/* CHANGES

26/3/95: replaced C library call to difftime() by my own mydifftime()
since SunOS doesn't include it in time.h. (ad@cs.mu.oz.au)

7/7/95: added graphics to style files. (ad@cs.mu.oz.au)

21/9/95: Each string used by the Entry and Section
structures is stored only once (in the strs_store[] array). So, 
rather than store a filename 5,000 times in the Entries list if there
are 5,000 files pointing to it, the filename string is stored
once in strs_store[] and there are 5,000 pointers to it. 
This means that curl can deal with larger sets of files.
This idea came from Jan Vitek (jvitek@cui.unige.ch).
Thanks Jan. (ad@cs.mu.oz.au)

MAXLINE has been increased to 1000 to deal with the large lines 
generated by keywords processing. There is also an error
message if the MAXLINE limit is exceeded. (ad@cs.mu.oz.au)

22/9/95: curl now allows two files to have the same name, but
the directories where they are located must be uniquely named
in that case. (ad@cs.mu.oz.au)

Fixed a bug in the update of the KEYSHTML file. (ad@cs.mu.oz.au)

24/10/95: Peter Martin <peterm@ind.tansu.com.au> found a bug
in free_section() to do with gcc's default assignment to pointers.


14/11/95: Pages now have a link to a SRCHHTML form for searching the 
document for pages containing keywords. This uses a CGI script
called sk, coded in C. (ad@cs.mu.oz.au)

A '*' against a name in the contents file denotes
that it is 'NEW'. This means that a 'NEW' picture will appear
against its entry in any contents lines in the pages. (ad@cs.mu.oz.au)

There is a link to a NEWHTML file which lists all the
NEW pages (those with a '*' against them in the contents file). The
list includes the title of each page and a link to each of them.
(ad@cs.mu.oz.au)

3/1/96. Modified find_body() to accept <body ... and </body>
(ad@cs.mu.oz.au)

25/3/96: Adonis El Fakih (eazl@gpd.plh.af.mil) suggested having icons
to represent no links. He used grayed icons, but I have used icons
with red crosses through them. write_link() has been modified
so that a No <word> link is replaced by a 'no link' image of the
icon. If the appropriate gif is not present, No <word> is used. 
A 'no_link' icon can be created using no_link.

*/


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>   /* used to check modification time of files */
#include <sys/stat.h>


#define ENTS	"prev_contents"     /* previous contents */
#define HDSTY	"dheader.sty"       /* default header style */
#define FTSTY	"dfooter.sty"       /* default footer style */
#define CNTSSTY	"dcnts.sty"         /* default contents style */

#define KEYSHTML    "keys.html"     /* key words file */
#define FIGSHTML    "figs.html"     /* figures file */
#define TBLSHTML    "tables.html"   /* tables file */
#define SRCHHTML    "search.html"   /* search form */
#define NEWHTML     "new.html"      /* list of NEW pages */

#define ARROWS      "arrows/"       /* the arrow icons directory */
#define NEWGIF      "new.gif"       /* picture representing 'NEW' */

#define NAMELEN	40        /* max length of a word */
#define MAXLINE	1000      /* max length of input line */
#define TOKLEN	41        /* max length of a token being parsed */
#define PATHLEN	100       /* max length of a word with a dir path */
#define LINKLEN	8         /* max length of link words:
            prev, home, next, up, keys, figs, tables, cnts, fcnts */
#define EKEYNO	15        /* max number of keys/figs/tbls in a single file */
#define READLNS	30        /* max lines read for keyword/string/TITLE */
#define STRSSIZE 27       /* 0-25 for letters; 26th for symbols */

typedef enum tokenkinds {NAME, NEWPG, LBRACK, RBRACK, EQUALS,
                         PATH, HEAD, FOOT, CNT, SCANEOF} Token;
enum entrykind {NEW, UNCHANGED, CHANGED};
enum stylekind {HEADER, FOOTER};
enum linkkind {PREV, HOME, NEXT, UP, KEYS, FIGS, TBLS, 
               SRCH, NEWPGS, INDEX, FINDEX, NONE};
enum refkind {ADD, DEL};


/* Strings used to store strings globally.
   Only one occurrence of any string is stored.
*/
typedef struct strnode {
  char *wd;
  struct strnode *next;
} Strnode;

typedef Strnode *Strings;


typedef struct siblings {
  Strings name;          /* eg foo */
  Strings fname;         /* from entry: eg "a/f.html" or "b.html"  */
  struct siblings *nextsib;
} Sibs;


/* Entry is a list holding the contents info. It is the 
   basis of the format of ENTS */
typedef struct entry {
  Strings name;           /* eg foo */
  Strings path;           /* from contents: eg "" or "b/" or "c/b/"  */
  enum entrykind ek;
  Strings dir;            /* from entry above: eg "" or "b" */
  int new;                /* flag for a NEW page */
  Strings prev, next, up, home, key, fig, tbl, srch, newpgs;
                        /* in relation to entry: eg a.html */
  Strings header, footer, indx;
                        /* from contents: eg "a/b/c.sty" or "c.sty" */
  int nokids;
  Sibs *kids;
  struct entry *nextentry;
} Entry;


/* Section is a tree d.s. which is used when the Contents file
   is first read in. It is converted into an Entry d.s. later */
typedef struct section {     
  Strings name;   /* see section defns */
  Strings path;
  Strings dir;
  int new;
  Strings home, key, fig, tbl, srch, newpgs;
  Strings header, footer, indx;  
  struct section *prev_page, *next_page;
  struct section *subs;
  struct section *parent;
} Section;


/* Keyref stores the references after a keyword */
typedef struct keyref {       /* in ascending order by nm */
  char *nm;        /* name of file: eg "a" */
  char *fullfn;    /* full file: eg "bar/a.html" */
  enum refkind rk;
  struct keyref *nextref;
} Keyref;

/* Key is the list d.s for the keywords */
typedef struct key {
  char *word;   /* eg "blue squares" in lowercase */
  Keyref *krs;
  struct key *nextkey;
} Key;


/* Entkey is briefly used when reading keys info from a file.
   This is then added to an on-going Keys d.s for all files */
typedef struct entkey {
  char *word;  /* eg "blue squares" in lowercase */
  enum refkind rk;
} Entkey;


/* Pic is used to hold the list of figures (and tables)
   mentioned in the html files. */
typedef struct pic {
  char *title;
  char *fullfn;
  int lineno;       /* used as a NAME anchor for the figure */
  struct pic *nextpic;
} Pic;



/* tokeniser globals */
char token_buffer[TOKLEN];
int buflen = 0;
Token curtok;

/* Global string storage: strs_store[] uses a simple
   form of hashing, storing a list of strings starting with 'a' (or 'A')
   in the first array cell, strings starting with 'b' in the second, 
   etc. The 27th cell is for strings that start with non-alphabetic 
   characters
 */
Strings strs_store[STRSSIZE];

Strings NIX;   /* This will point to the string "" in strs_store */
Strings KH;    /* Will point to KEYSHTML string */
Strings FH;    /* Will point to FIGSHTML string */
Strings TH;    /* Will point to TBLSHTML string */
Strings SH;    /* Will point to the SRCHHTML string */
Strings PH;    /* Will point to the NEWHTML string */


/* top-level and tokeniser functions */
FILE *process_args(int argc, char *argv[], int *sflag);
void get_tok(FILE *fp);
Token scanner(FILE *fp);
void buffer_char(char c);
void print_buffer(void);
void lexical_error(char c);
Token check_reserved(void);
void print_token(Token t);
void syntax_error(Token actual, Token expected);

/* parsing functions */
Section *page(FILE *fp, Section *parent, Strings path, Strings dir,
                 Strings header, Strings footer, Strings indx);
Section *new_pg(Section *par, Strings path, Strings dir,
                  Strings header, Strings footer, Strings indx);
void style(FILE *fp, Section *ppg, Strings *npath, Strings *ndir,
             Strings *nheader, Strings *nfooter, Strings *nindx);

char *sassign(char *s);
char *extend_path(char *oldpath, char *ndir);
Strings str_extend_path(char *oldpath, char *ndir);
char *htmlending(char *nm);
Strings str_htmlending(char *nm);
char *add_path(char *path, char *fnm);
Strings str_add_path(char *path, char *fnm);
char *add_path_html(char *path, char *fnm);

void print_contents(Section *cnts);

/* build section and entries d.s */
Entry *build(Section *cnts);
Entry *build_entry(Section *cnts);
Sibs *build_kids(Section *sub, int *n);
void build_entries(Section *sec, Entry **curr);
void check_dups(Sibs *sbs, char *path);
void nodup_kids(char *name, char *path, Sibs *sbs);

void free_section(Section *s);

/* some print utilities for debugging the Entry d.s */
void print_entries(Entry *es);
void print_ek(enum entrykind ek);
void print_sibs(Sibs *s);

/* load old entries from ENTS and find differences with new entries */
Entry *load_entries(FILE *fp);
int scan_entry(Entry *e, FILE *fp);
Strings scan_strsbld(char *s);
void extract_kids(FILE *fp, Sibs *s, int no);
void find_changes(Entry *new, Entry **pold);
void find_change(Entry *new, Entry **pold);
int same_contents(Entry *new, Entry *old);
int same_sibs(Sibs *nkids, Sibs *okids);
void free_entry(Entry *e);

/* update the files mentioned in the Contents */
void update_entries(Entry *new, int silent, 
                      Key **pkeys, Pic **pfigs, Pic **ptbls);
void recent_edit(char *htmlnm, Entry *ent, Entry *top, 
                 int silent, Key **pkeys, Pic **pfigs, Pic **ptbls);
void update_html(char *htmlnm, Entry *ent, Entry *top);
void write_style(enum stylekind sk, FILE *fp, FILE *sfp, Entry *et);
int get_link(FILE *fp, char word[], enum linkkind *plk);
int get_gifnm(FILE *sfp, Entry *et, char gifnm[], char no_nm[]);
int check_gifnm(char *name);
void path_to_gifs(char *home, char path[]);
void write_link(FILE *fp, char *word, char *gifnm, char *no_nm,
                         enum linkkind lk, Entry *et);
void write_cnt(FILE *hfp, Entry *et, Entry *top);
void write_fcnts(FILE *hfp, Entry *et, Entry *top);
int linkline(char line[], char lstart[], char lend[]);
void write_cntsline(FILE *fp, char *pathhome, char *pathfnm, 
                    Sibs *kid, Entry *et, char *lnstart, char *lnend);
void read_title(Entry *et, char *name, char string[]);
int find_title(char *ffn, char line[], char string[]);
int start_title(char line[], int i);
int end_title(char line[], int i);
void next_cnt(FILE *fp, char *pathhome, Sibs *kdy, char *path, Entry *top,
                                    char *lnstart, char *lnend);
Entry *find_entry(Sibs *kid, Entry *top);
int last_header(FILE *fp);
int curl_temp(char *htmlnm, FILE *fromfp, FILE *tofp, FILE *hdsfp,
                FILE *ftsfp, Entry *ent, Entry *top);
void find_body(char *nm, FILE *fp, int *start, int *end);
int write_middle(char *nm, FILE *tfp, FILE *hfp);
int same_chars(char line[], int index, char pat[]);
void strip_pic(char line[], char nline[]);

void save_entries(Entry *new);
char *star(char *s);
void announce_old(Entry *old);

/* collect keys from files */
void collect_keys(char *htmlnm, Entry *ent, Key **pkeys);
int read_ekeys(char *ffn, Entkey ekeys[], int *keyno);
void add_to_keys(Entkey ek, char *nm, char *ffn, Key **pkeys);
Key *new_key(Entkey ek, char *nm, char *ffn);
Keyref *new_ref(char *nm, char *ffn, enum refkind rk);
void add_ref(Keyref **krs, Entkey ek, char *nm, char *ffn);
int find_ekey(char line[], Entkey keys[], int *keyno);
void make_lower(char word[]);

/* print utilities for debugging Key d.s */
void print_keys(Key *ks);
void print_keyref(Keyref *krs);
void print_refkind(enum refkind rk);

/* update key info in KEYSHTML */
void update_kfile(Key *keys);
void modify_kf(Key *keys, FILE *kfp, FILE *tkfp, char *line, int prevch);
void hrefs_inserts(char *line, Key *keys);
void delete_lineref(char *line, int i);
void insert_ref(Keyref *refs, char *line, int *pi, char *kword);
void add_string(char *line, int *pi, char *s);
void key_inserts(Key *keys, FILE *tkfp);
void copy_to_entries(FILE *kfp, FILE *tkfp, int prevch, char *line, int ch);
int char_comment(char *line, int ch);
int find_key(char *line, char *key);
int html_cmp(char *line, int i, char *nm);
int ffn_cmp(char *line, int i, char *refffn);
int a_or_p(char *line, int i);

void free_key(Key *k);
void free_krefs(Keyref *kr);

/* Collect Pic information */
void collect_pics(char *ffn, Entry *ent, Pic **pfigs, Pic **ptbls);
int find_pics(char *ffn, char line[], int lineno, char *piclabel,
          char *picword, Entry *ent, FILE *tfp, Pic **pp);
void build_picline(char line[], int i, Entry *ent, char *picword,
                              int lineno, char nline[]);
void itoa(int n, char s[]);
void add_to_pics(char *ffn, char *desc, int lineno, Pic **pp);
Pic *new_pic(char *desc, char *ffn, int lineno);

/* Update Pic information in FIGSHTML/TBLSHTML */
void mod_picfile(Pic *pics, char *ftitle, char *picfn);
void delete_oldpics(Pic *pics, char line[], FILE *tfp);
void insert_newpics(Pic *pics, FILE *tfp);

/* print utility for debugging pic d.s */
void print_pics(Pic *pcs);

/* Update new.html */
void mod_newfile(Entry *top);
void insert_newpgs(Entry *el, FILE *tfp);

/* redefine difftime() since SunOS doesn't have it! */
double mydifftime(time_t t1, time_t t0);

/* Strings manipulation */
Strings strsbld(char *s);
int strs_hash(char *s);
Strings strs_insert(Strings *list, char *s);
Strings mk_strs(char *s);
void init_strs(void);
void print_strs(void);
void print_str(Strings list);


/* xxx */


int main(int argc, char *argv[])
/* Parse the Contents file, building a Section d.s called cnts.
   Converts cnts into an Entries d.s., new_entries.
   Read in old_entries from the previous contents file ENTS.
   Find the changes between old_entries and new_entries.
   Use the changes to update the files referred to in Contents.
   Also update the key words file KEYSHTML, the figures file
   FIGSHTML, and the tables file TBLSHTML.
   Save the new_entries in ENTS for next time.
*/
{
  Entry *new_entries, *old_entries;
  Section *cnts;
  FILE *cfp, *efp;
  int silent;    /* if set, don't say a file is unchanged */
  Key *keys = NULL;
  Pic *figs, *tbls;

  init_strs(); 

  figs = NULL; tbls = NULL;
  if ((cfp = process_args(argc, argv, &silent)) != NULL) {
    get_tok(cfp);
    cnts = page(cfp, NULL, NIX, NIX, strsbld(HDSTY), 
                              strsbld(FTSTY), strsbld(CNTSSTY));
    fclose(cfp);
    if (curtok != SCANEOF)
      syntax_error(curtok, SCANEOF);
    else {
   /* printf("\n\ncnts is:\n");
      print_contents(cnts); */
      new_entries = build(cnts);  
      free_section(cnts);
   /* printf("\n\nnew_entries is:\n");
      print_entries(new_entries); */
      if ((efp = fopen(ENTS,"r")) != NULL) {
        printf("\nComparing with previous contents in %s\n", ENTS);
        old_entries = load_entries(efp);
   /*   printf("\n\n1. old_entries is:\n");
        print_entries(old_entries);  */
        fclose(efp);
        find_changes(new_entries, &old_entries);
   /*   printf("\n\n2. old_entries is:\n");
        print_entries(old_entries); */
      }
      else
        old_entries = NULL;
      update_entries(new_entries, silent, &keys, &figs, &tbls);
      update_kfile(keys);
      putchar('\n');
   /* print_strs(); */
   /* printf("\nfigs is:\n");
      print_pics(figs); */
      mod_picfile(figs, "Figures", FIGSHTML);
   /* printf("\ntbls is:\n");
      print_pics(tbls); */
      mod_picfile(tbls, "Tables", TBLSHTML);
      mod_newfile(new_entries);
      save_entries(new_entries);
      announce_old(old_entries);
    }
  }
  return 0;
}


FILE *process_args(int argc, char *argv[], int *silent)
{
  FILE *fp;

  *silent = 0;
  if (argc == 2) {
    if ((fp = fopen(argv[1],"r")) == NULL) {
      printf("curl: cannot open %s\n", argv[1]);
      printf("Usage: curl [-s] file\n");
      return NULL;
    }
    else
      return fp;
  }
  if (argc == 3) {
    if (strcmp(argv[1], "-s") == 0) {
      *silent = 1;
    }
    else {
      printf("Usage: curl [-s] file\n");
      printf("Illegal Option: %s being ignored\n", argv[1]);
    }
    if ((fp = fopen(argv[2],"r")) == NULL) {
      printf("Cannot open %s\n", argv[2]);
      printf("Usage: curl [-s] file\n");
      return NULL;
    }
    else
      return fp;
  }
  else {
    printf("Wrong number of arguments\n");
    printf("Usage: curl [-s] file\n");
    return NULL;
  }
}


/* Tokeniser functions used on the Contents file */

/* Typical contents file:
        f1 {
          f2 *
          f3 {
            path=d head = t1   // a comment to eoln
            foot=b1  cnts = i
              f4 *
              f5
          }
        }
The settings are optional but must be that order. The
style names are for files ending in .sty

Every filename can be followed by a '*', which indicates
that it is NEW.
*/

void get_tok(FILE *fp)
{
  curtok = scanner(fp);
  /* printf("Token buffer: "); print_buffer(); putchar('\n');
  printf("Curtok is: "); print_token(curtok); printf("\n\n"); */
}

Token scanner(FILE *fp)
{ 
  int inchar, c;

  buflen = 0;
  if (feof(fp))
    return SCANEOF;
  while ((inchar = getc(fp)) != EOF) {
    if (isspace(inchar))
      continue;    /* do nothing */
    else if (isalpha(inchar)) {      
           /* id := letter ( letter | digit | underscore | dash )*  */
      buffer_char(inchar);
      for (c = getc(fp); (isalnum(c) || c == '_' || c == '-'); c = getc(fp))
        buffer_char(c);
      ungetc(c, fp);
      return check_reserved();
    }
    else if (inchar == '{')
      return LBRACK;
    else if (inchar == '}')
      return RBRACK;
    else if (inchar == '=')
      return EQUALS;
    else if (inchar == '*')
      return NEWPG;
    else if (inchar == '/') {
         /* is this a // comment line? */
       c = getc(fp);
       if (c == '/')
         do
           inchar = getc(fp);
         while (inchar != '\n');
       else
         lexical_error(c);
    }
    else
      lexical_error(inchar);
  }
  if (feof(fp))
    return SCANEOF;
}

void buffer_char(char c)    /* cast of c */
{
  if (buflen == (TOKLEN-1)) {
    printf("Cannot add %c due to token length limit of %d. Token is: ", c, TOKLEN-1);
    print_buffer();
    putchar('\n');
  }
  else
    token_buffer[buflen++] = c;
}

void print_buffer(void)
{ 
  int i;
  for (i=0; i < buflen; i++)
    putchar(token_buffer[i]);
}

void lexical_error(char c)   /* cast of c */
{
  printf("In contents file, do not recognise: %c\n", c);
  exit(1);
}

Token check_reserved(void)
{
  token_buffer[buflen] = '\0';     /* make into string array */
  if (strcmp(token_buffer, "path") == 0)
    return PATH;
  else if (strcmp(token_buffer, "head") == 0)
    return HEAD;
  else if (strcmp(token_buffer, "foot") == 0)
    return FOOT;
  else if (strcmp(token_buffer, "cnts") == 0)
    return CNT;
  else
    return NAME;
}

void print_token(Token t)
{
  if (t == LBRACK)
    putchar('{');
  else if (t == RBRACK)
    putchar('}');
  else if (t == EQUALS)
    putchar('=');
  else if (t == NEWPG)
    putchar('*');
  else if (t == PATH)
    printf("path");
  else if (t == HEAD)
    printf("head");
  else if (t == FOOT)
    printf("foot");
  else if (t == CNT)
    printf("cnts");
  else if (t == SCANEOF)
    printf("eof");
  else
    printf("$token$");
}

void syntax_error(Token actual, Token expected)
{
  printf("In the contents file, the actual token is: ");
  if (actual == NAME)
    printf("%s", token_buffer);
  else
    print_token(actual);
  putchar('\n');
  printf("The expected token is: ");
  if (expected == NAME)
    printf("a name");
  else
    print_token(expected);
  putchar('\n');
  exit(1);
}


/* syntax analysis: 
   Contents file tokens built into Section tree */

Section *page(FILE *fp, Section *parent, Strings path, Strings dir,
                 Strings header, Strings footer, Strings indx)
/* BNF  page ::= name [*] [ '{' style page page* '}' ]  
   A page is a filename mentioned in the Contents 
*/
{
  Strings npath, ndir, nheader, nfooter, nindx;
  Section *npage, *child;

  npage = NULL;
  if (curtok == NAME) {
    npage = new_pg(parent, path, dir, header, footer, indx);
    get_tok(fp);
    if (curtok == NEWPG) {
      npage->new = 1;
      get_tok(fp);
    }
    else
      npage->new = 0;
    if (curtok == LBRACK) {
      get_tok(fp);
      style(fp, npage, &npath, &ndir, &nheader, &nfooter, &nindx);
      npage->subs = page(fp, npage, npath, ndir, nheader, nfooter, nindx);
      npage->subs->prev_page = NULL;
      child = npage->subs;
      while (curtok != RBRACK) {
        child->next_page =
            page(fp, npage, npath, ndir, nheader, nfooter, nindx);
        child->next_page->prev_page = child;
        child = child->next_page;
      }
      child->next_page = NULL;
      get_tok(fp);
    }
  }
  else
    syntax_error(curtok, NAME);
  return npage;
}


Section *new_pg(Section *par, Strings path, Strings dir,
                  Strings header, Strings footer, Strings indx)
/* Build a single Section node, corresponding 
   to a 'page' in the contents */
{
  Section *npage;

  npage = (Section *) malloc(sizeof(Section));
  npage->name = strsbld(token_buffer);
  npage->path = path;
  npage->dir = dir;
  npage->header = header;
  npage->footer = footer;
  npage->indx = indx;
  npage->prev_page = NULL;
  npage->next_page = NULL;
  npage->subs = NULL;
  npage->parent = par;
  if (par == NULL) {
    npage->home = str_htmlending(npage->name->wd);   /* I'm the home page */
    npage->key = KH;           /* key file in same dir */
    npage->fig = FH;           /* figs file in same dir */
    npage->tbl = TH;           /* tbls file in same dir */
    npage->srch = SH;          /* search form in same dir */
    npage->newpgs = PH;        /* new file in same dir */
  }
  else if (dir == NIX) {        /* same dir as parent */
    npage->home = par->home;
    npage->key = par->key;
    npage->fig = par->fig;
    npage->tbl = par->tbl;
    npage->srch = par->srch;
    npage->newpgs = par->newpgs;
  }
  else {      /* one dir below parent */
    npage->home = str_add_path("../", par->home->wd);
    npage->key =  str_add_path("../", par->key->wd);
    npage->fig =  str_add_path("../", par->fig->wd);
    npage->tbl =  str_add_path("../", par->tbl->wd);
    npage->srch = str_add_path("../", par->srch->wd);
    npage->newpgs = str_add_path("../", par->newpgs->wd);
  }
  return npage;
}


void style(FILE *fp, Section *ppg, Strings *npath, Strings *ndir,
             Strings *nheader, Strings *nfooter, Strings *nindx)

/* BNF  style ::= [ 'path' '=' name ]  [ 'head' '=' name ]
                  [ 'foot' '=' name ]   [ 'cnts' '=' name ]
*/
{
  char word[NAMELEN];

  if (curtok == PATH) {
    get_tok(fp);
    if (curtok == EQUALS) {
      get_tok(fp);
      if (curtok == NAME) {
        *ndir = strsbld(token_buffer);
        *npath = str_extend_path(ppg->path->wd, token_buffer);
        get_tok(fp);
      }
      else syntax_error(curtok, NAME);
    }
    else syntax_error(curtok, EQUALS);
  }
  else {
    *ndir = NIX;
    *npath = ppg->path;
  }
  
  if (curtok == HEAD) {
    get_tok(fp);
    if (curtok == EQUALS) {
      get_tok(fp);
      if (curtok == NAME) {
        sprintf(word, "%s%s.sty", (*npath)->wd, token_buffer);
        *nheader = strsbld(word);
        get_tok(fp);
      }
      else syntax_error(curtok, NAME);
    }
    else syntax_error(curtok, EQUALS);
  }
  else *nheader = ppg->header;

  if (curtok == FOOT) {
    get_tok(fp);
    if (curtok == EQUALS) {
      get_tok(fp);
      if (curtok == NAME) {
        sprintf(word, "%s%s.sty", (*npath)->wd, token_buffer);
        *nfooter = strsbld(word);
        get_tok(fp);
      }
      else syntax_error(curtok, NAME);
    }
    else syntax_error(curtok, EQUALS);
  }
  else *nfooter = ppg->footer;

  if (curtok == CNT) {
    get_tok(fp);
    if (curtok == EQUALS) {
      get_tok(fp);
      if (curtok == NAME) {
        sprintf(word, "%s%s.sty", (*npath)->wd, token_buffer);
        *nindx = strsbld(word);
        get_tok(fp);
      }
      else syntax_error(curtok, NAME);
    }
    else syntax_error(curtok, EQUALS);
  }
  else *nindx = ppg->indx;

}


char *sassign(char *s)
{
  char *ns;
  ns = (char *) malloc(sizeof(char)*(strlen(s)+1));
  strcpy(ns, s);
  return ns;
}



char *extend_path(char *oldpath, char *ndir)
/* Paths are "", "bar/", "bar/foo/", etc */
{ 
  char *ns;
  ns = (char *)malloc(sizeof(char)*(strlen(oldpath)+strlen(ndir)+2));

  if (strcmp(ndir, "") == 0)
    strcpy(ns, oldpath);
  else
    sprintf(ns, "%s%s/", oldpath, ndir);
  return ns;
}

Strings str_extend_path(char *oldpath, char *ndir)
/* Combines string generation and storage in strs_store[] */
{
  Strings s;
  char *temp;
  temp = extend_path(oldpath, ndir);
  s = strsbld(temp);
  free(temp);
  return s;
}


char *htmlending(char *nm)    
/* Add .html to end of nm */
{
  char *ns;
  ns = (char *) malloc(sizeof(char)*(strlen(nm)+6));
  sprintf(ns, "%s.html", nm);
  return ns;
}

Strings str_htmlending(char *nm)
/* Combines string generation and storage in strs_store[] */
{
  Strings s;
  char *temp;
  temp = htmlending(nm);
  s = strsbld(temp);
  free(temp);
  return s;
}


char *add_path(char *path, char *fnm)
/* Path is "" or "../" or "../../", etc */
{
  char *ns;
  ns = (char *) malloc(sizeof(char) * (strlen(path) + strlen(fnm) + 1));
  sprintf(ns, "%s%s", path, fnm);
  return ns;
}

Strings str_add_path(char *path, char *fnm)
/* Combines string generation and storage in strs_store[] */
{
  Strings s;
  char *temp;
  temp = add_path(path, fnm);
  s = strsbld(temp);
  free(temp);
  return s;
}


char *add_path_html(char *path, char *fnm)
/* ns is assigned path + fnm + .html */
{
  char *ns;
  ns = (char *) malloc(sizeof(char) * (strlen(path) + strlen(fnm) + 6));
  sprintf(ns, "%s%s.html", path, fnm);
  return ns;
}


void print_contents(Section *cnts)
/* Debugging utility for printing Contents */
{
  if (cnts == NULL)
    printf("NULL\n");
  else {
    printf("Node Values\n\n");
    printf("  name is: %s\n", cnts->name->wd);
    printf("  path is: %s\n", cnts->path->wd);
    printf("  dir is: %s\n", cnts->dir->wd);
    printf("  new is: %d\n", cnts->new);
    printf("  home is: %s\n", cnts->home->wd);
    printf("  key is: %s\n", cnts->key->wd);
    printf("  fig is: %s\n", cnts->fig->wd);
    printf("  tbl is: %s\n", cnts->tbl->wd);
    printf("  srch is: %s\n", cnts->srch->wd);
    printf("  newpgs is: %s\n", cnts->newpgs->wd);
    printf("  header is: %s\n", cnts->header->wd);
    printf("  footer is: %s\n", cnts->footer->wd);
    printf("  indx is: %s\n", cnts->indx->wd);
    if (cnts->prev_page == NULL)
      printf("  Previous page is NULL\n");
    else
      printf("  Previous page is %s\n", cnts->prev_page->name->wd);
    if (cnts->parent == NULL)
      printf("  Parent page is NULL\n");
    else
      printf("  Parent page is %s\n", cnts->parent->name->wd);
    printf("  Visit Children of %s ***\n", cnts->name->wd);
    print_contents(cnts->subs);
    printf("\n\n");
    printf("  Next page after %s\n", cnts->name->wd);
    print_contents(cnts->next_page);
  }
}


/* Build an entries list from the contents */

Entry *build(Section *cnts)
{
  Entry *top, *curr;
  Section *s;

  top = build_entry(cnts);
  s = cnts->subs;
  curr = top;
  while (s != NULL) {
    build_entries(s, &curr);
    s = s->next_page;
  }
  return top;
}


Entry *build_entry(Section *cnts)
/* Build a single Entry node */
{
  char uppath[NAMELEN];
  Entry *e;

  if (cnts == NULL)
    return NULL;
  else {
    e = (Entry *) malloc(sizeof(Entry));
    e->name = cnts->name;
    e->path = cnts->path;
    e->ek = NEW;
    e->dir = cnts->dir;
    e->new = cnts->new;
    e->header = cnts->header;
    e->footer = cnts->footer;
    e->indx = cnts->indx;
    e->home = cnts->home;
    e->key = cnts->key;
    e->fig = cnts->fig;
    e->tbl = cnts->tbl;
    e->srch = cnts->srch;
    e->newpgs = cnts->newpgs;
    if (cnts->parent == NULL)    /* I'm the home page */
      e->up = str_htmlending(cnts->name->wd);
    else if (e->dir == NIX)    /* same dir as parent */
      e->up = str_htmlending(cnts->parent->name->wd);
    else {
      sprintf(uppath, "../%s.html", cnts->parent->name->wd);
      e->up = strsbld(uppath);
    }
    if (cnts->prev_page == NULL)
      e->prev = NIX;
    else
      e->prev = str_htmlending(cnts->prev_page->name->wd);
    if (cnts->next_page == NULL)
      e->next = NIX;
    else
      e->next= str_htmlending(cnts->next_page->name->wd);
    e->kids = build_kids(cnts->subs, &(e->nokids));
    check_dups(e->kids, e->path->wd);
    e->nextentry = NULL;

    return e;
  }
}


Sibs *build_kids(Section *sub, int *n)
/* Build the children of an Entry */
{
  Sibs *kid, *start;
  char newname[NAMELEN];

  if (sub == NULL) {
    *n = 0;
    return NULL;
  }
  else {
    kid = (Sibs *) malloc(sizeof(Sibs));
    start = kid;
    *n = 1;
    while (sub != NULL) {
      kid->name = sub->name;
      if (sub->dir == NIX)      /* sub in same directory as parent */
        kid->fname = str_htmlending(sub->name->wd);
      else {
        sprintf(newname, "%s/%s.html", sub->dir->wd, sub->name->wd);
        kid->fname = strsbld(newname);
      }
      if (sub->next_page != NULL) {
        kid->nextsib = (Sibs *) malloc(sizeof(Sibs));
        (*n)++;
        kid = kid->nextsib;
      }
      else
        kid->nextsib = NULL;
      sub = sub->next_page;
    }
    return start;
  }
}


void build_entries(Section *sec, Entry **pcurr)
{
  Section *s;

  (*pcurr)->nextentry = build_entry(sec);
  s = sec->subs;
  (*pcurr) = (*pcurr)->nextentry;
  while (s != NULL) {
    build_entries(s, pcurr);
    s = s->next_page;
  }
}


/* free up the Section d.s */
void free_section(Section *s)
{
  Section *curr, *next;

  curr = s;         /* correction; spotted by Peter Martin */
  while (curr != NULL) {
    next = curr->subs;
    free(curr);
    curr = next;
  }
}


void check_dups(Sibs *sbs, char *path)
/* Report all duplicate siblings. */
{
  Sibs *s;

  s = sbs;
  while (s != NULL) {
    nodup_kids(s->name->wd, path, s->nextsib);
    s = s->nextsib;
  }
}


void nodup_kids(char *name, char *path, Sibs *sbs)
/* Search along Sibs list for repetitions of name, and report them. */
{
  Sibs *s;
  char *ffn;

  s = sbs;
  while (s != NULL) {
    if (strcmp(name, s->name->wd) == 0) {
      ffn= add_path(path, s->fname->wd);
      printf("Found duplicate in contents file: %s\n", ffn);
      free(ffn);
    }
    s = s->nextsib;
  }
}


/* Debugging utilities for printing out Entries */

void print_entries(Entry *es)
{
  if (es == NULL)
    printf("NULL\n");
  else {
    printf("Node Values\n\n");
    printf("  name is: %s\n", es->name->wd);
    printf("  path is: %s\n", es->path->wd);
    printf("  entry kind is: "); print_ek(es->ek); putchar('\n');
    printf("  dir is: %s\n", es->dir->wd);
    printf("  new is: %d\n", es->new);
    printf("  prev is: %s\n", es->prev->wd);
    printf("  next is: %s\n", es->next->wd);
    printf("  up is: %s\n", es->up->wd);
    printf("  home is: %s\n", es->home->wd);
    printf("  key is: %s\n", es->key->wd);
    printf("  fig is: %s\n", es->fig->wd);
    printf("  tbl is: %s\n", es->tbl->wd);
    printf("  srch is: %s\n", es->srch->wd);
    printf("  newpgs is: %s\n", es->newpgs->wd);
    printf("  header is: %s\n", es->header->wd);
    printf("  footer is: %s\n", es->footer->wd);
    printf("  indx is: %s\n", es->indx->wd);
    printf("  nokids is: %d\n", es->nokids);
    if (es->nokids > 0) {
      printf("  Visit Children of %s ***\n", es->name->wd);
      print_sibs(es->kids);
      printf("\n\n");
    }
    printf("  Next entry after %s\n", es->name->wd);
    print_entries(es->nextentry);
  }
}

void print_ek(enum entrykind ek)
{
  switch(ek) {
    case NEW: printf("NEW"); break;
    case UNCHANGED: printf("UNCHANGED"); break;
    case CHANGED: printf("CHANGED"); break;
  }
}

void print_sibs(Sibs *s)
{
  Sibs *temp;
  temp = s;
  while (temp != NULL) {
    printf("    Sib name: %s\n", temp->name->wd);
    printf("    Sib fname: %s\n", temp->fname->wd);
    temp = temp->nextsib;
  }
}


/* Load old entries from ENTS and find differences with new entries */

Entry *load_entries(FILE *fp)
{
  Entry *fst, *tail;
  fst = (Entry *) malloc(sizeof(Entry));
  scan_entry(fst, fp);
  tail = fst;
  tail->nextentry = (Entry *) malloc(sizeof(Entry));
  while (scan_entry(tail->nextentry, fp) == 1) {
    tail = tail->nextentry;
    tail->nextentry = (Entry *) malloc(sizeof(Entry));
  }
  free(tail->nextentry);
  tail->nextentry = NULL;
  return fst;
}


int scan_entry(Entry *e, FILE *fp)
/* Read in a line from ENTS (everything is a string, with 
   empty strings represented as * in the file.
   Each line corresponds to an Entry node.  */
{
  int nokids, new;
  char name[NAMELEN], path[PATHLEN], dir[NAMELEN], prev[NAMELEN],
       next[NAMELEN], up[PATHLEN], home[PATHLEN], key[PATHLEN],
       fig[PATHLEN], tbl[PATHLEN], srch[PATHLEN], newpgs[PATHLEN],
       header[PATHLEN], footer[PATHLEN], indx[PATHLEN];

  if (fscanf(fp, "%s %s %s %d %s %s %s %s %s %s %s %s %s %s %s %s %d",
                name, path, dir, &new, prev, next, up, home,
                key, fig, tbl, srch, newpgs, header, footer, 
                indx, &nokids) == 17) {
    e->name = scan_strsbld(name);
    e->path = scan_strsbld(path);
    e->dir  = scan_strsbld(dir);
    e->new = new;
    e->prev = scan_strsbld(prev);
    e->next = scan_strsbld(next);
    e->up   = scan_strsbld(up);
    e->home = scan_strsbld(home);
    e->key  = scan_strsbld(key);
    e->fig  = scan_strsbld(fig);
    e->tbl  = scan_strsbld(tbl);
    e->srch  = scan_strsbld(srch);
    e->newpgs  = scan_strsbld(newpgs);
    e->header = scan_strsbld(header);
    e->footer = scan_strsbld(footer);
    e->indx = scan_strsbld(indx);
    e->nokids = nokids;
    if (e->nokids > 0) {
      e->kids = (Sibs *) malloc(sizeof(Sibs));
      extract_kids(fp, e->kids, e->nokids);
    }
    else e->kids = NULL;
    return 1;
  }
  else return 0;
}

Strings scan_strsbld(char *s)
/* Empty string is stored as "*"; replace by "" */
{ 
  if (strcmp(s,"*") == 0)
    return NIX;
  else 
    return strsbld(s);
}


void extract_kids(FILE *fp, Sibs *s, int no)
/* Read in an Entry's children from ENTS */
{
  char word[NAMELEN];
  Sibs *temp;

  fscanf(fp, "%s", word);
  s->name = strsbld(word);
  fscanf(fp, "%s", word);
  s->fname = strsbld(word);
  temp = s;
  while (--no > 0) {
    temp->nextsib = (Sibs *) malloc(sizeof(Sibs));
    temp = temp->nextsib;
    fscanf(fp, "%s", word);
    temp->name = strsbld(word);
    fscanf(fp, "%s", word);
    temp->fname = strsbld(word);
  }
  temp->nextsib = NULL;
}


/* Find the differences between the old entries and the new ones */
/* Old entries come from ENTS; new entries built from Contents file */

void find_changes(Entry *new, Entry **pold)
{
  while (new != NULL) {
    find_change(new, pold);
    new = new->nextentry;
  }
}

void find_change(Entry *new, Entry **pold)
{
  Entry *oldes, *prev;
  int found = 0;

  oldes = *pold; prev = NULL;
  while ((oldes != NULL) && (!found)) {
    if ((new->name == oldes->name) && (new->path == oldes->path)) {
      found = 1;
      if (same_contents(new,oldes) == 1)
        new->ek = UNCHANGED;
      else
        new->ek = CHANGED;
      if (prev == NULL) {   /* at start of entries */
        oldes = oldes->nextentry;
        free_entry(*pold);
        *pold = oldes;
      }
      else {
        prev->nextentry = oldes->nextentry;
        free_entry(oldes);
        oldes = NULL;
      }
    }
    else {
      prev = oldes;
      oldes = oldes->nextentry;
    }
  }
}


int same_contents(Entry *new, Entry *old)
/* Compare each file of old Entry node and new Entry node */
{
  return  (new->new == old->new) &&
          (new->prev == old->prev) && 
          (new->next == old->next) &&
          (new->up == old->up) && 
          (new->home == old->home) && 
          (new->key == old->key) && 
          (new->fig == old->fig) &&
          (new->tbl == old->tbl) &&
          (new->srch == old->srch) &&
          (new->newpgs == old->newpgs) &&
          (new->header == old->header) &&
          (new->footer == old->footer) &&
          (new->indx == old->indx) && 
          (new->nokids == old->nokids) && 
          (same_sibs(new->kids, old->kids) == 1) ;
}


int same_sibs(Sibs *nkids, Sibs *okids)
{
  Sibs *nks, *oks;
  nks = nkids, oks = okids;
  while ((nks != NULL) && (oks != NULL))
    if ((nks->fname != oks->fname) || (nks->name != oks->name))
      return 0;
    else {
      nks = nks->nextsib;
      oks = oks->nextsib;
    }
  if ((nks == NULL) && (oks == NULL))
    return 1;
  else
    return 0;
}


void free_entry(Entry *e)
/* Delete old_entries node which is not required */
{
  Sibs *s;

  if (e !=NULL) {
    while (e->kids != NULL) {
      s = e->kids;
      e->kids = e->kids->nextsib;
      free(s);
    }
   free(e);
  }
}

/* Update the files referred to in the Contents file
   based on the changes detected between new_entries and old_entries */

void update_entries(Entry *ent, int silent, 
                      Key **pkeys, Pic **pfigs, Pic **ptbls)

/* Uses the ek field of each Entry node which records the
   change status. */
/* New html files will have links added; html files with old 
   links will have them renewed. Links are the header, footer
   and any cnts/fcnts bits in the text. */
/* UNCHANGED files will still be updated *if* they have been
   modified since ENTS was last modified. */
/* The building of KEYSHTML/FIGSHTML/TBLSHTML are separated from the
   manipulation of the files. */
/* If silent is set then unchanged files are not reported. */
{
  Entry *top;
  char *htmlnm;

  top = ent;
  while (ent != NULL) {
    htmlnm = add_path_html(ent->path->wd, ent->name->wd);
    switch (ent->ek) {
      case NEW: case CHANGED:
        update_html(htmlnm, ent, top);
        collect_keys(htmlnm, ent, pkeys);
        collect_pics(htmlnm, ent, pfigs, ptbls);
        break;
      case UNCHANGED:
        recent_edit(htmlnm, ent, top, silent, pkeys, pfigs, ptbls);
        break;
      default:
        printf("\nIncorrect entry kind in %s%s\n", 
                              ent->path->wd, ent->name->wd);
        break;
    }
    free(htmlnm);
    ent = ent->nextentry;
  }
}


void recent_edit(char *htmlnm, Entry * ent, Entry *top, 
                 int silent, Key **pkeys, Pic **pfigs, Pic **ptbls)
/* Only reached if the Entry node for the file records
   UNCHANGED. However, the file may need updating if keys or
   contents information has been altered. */
/* If the HTML file has been modified since ENTS then it
   is updated. */
/* If silent is set then do not report an unchanged file. */

{
  struct stat entsb, filesb;

  if (stat(ENTS, &entsb) == 0) {
    if (stat(htmlnm, &filesb) == 0) {
      if (mydifftime(filesb.st_mtime, entsb.st_mtime) > 0) {
        /* printf("\n%s has been recently edited", htmlnm); */
        update_html(htmlnm, ent, top);
        collect_keys(htmlnm, ent, pkeys);
        collect_pics(htmlnm, ent, pfigs, ptbls);
      }
      else {       /* html file is unchanged */
        if (silent == 0)
          printf("\n%s is unchanged\n", htmlnm);
      }
    }
  }
  else
    printf("Cannot read stat structure for %s\n", ENTS);
}


void update_html(char *htmlnm, Entry *ent, Entry *top)
/* If the html file has not been processed before then
   links are added. Use last_header to check this. */
/* If the file has been processed before then its links 
   are 'refreshed'.
   This consists of stripping out the old header and footer
   and regenerating them. Similarly the old cnts/fcnts is
   removed and regenerated. */
{
  char tmpnm[PATHLEN];
  FILE *htmlfp, *hdstylfp, *ftstylfp, *tempfp;
  int last;

  if ((hdstylfp = fopen(ent->header->wd,"r")) == NULL) {
    printf("\nHeader style file %s does not exist\n", ent->header->wd);
    return;
  }
  if ((ftstylfp = fopen(ent->footer->wd,"r")) == NULL) {
    printf("\nFooter style file %s does not exist\n", ent->footer->wd);
    return;
  }

  if ((htmlfp = fopen(htmlnm,"r")) != NULL) {     /* check .html */
    strcpy(tmpnm, htmlnm);
    strcat(tmpnm, "tmp");
    if ((tempfp = fopen(tmpnm,"w")) == NULL) {
      printf("\nCannot open temporary file for %s\n", htmlnm);
      remove(tmpnm);
      fclose(htmlfp);
    }
    else {
      last = last_header(htmlfp);
      if (last == -1) {      /* not a html file with curl links */
        printf("\n%s not previously processed by curl\n", htmlnm);
        if (curl_temp(htmlnm, htmlfp, tempfp, hdstylfp, ftstylfp, 
                                             ent, top) == 1) {
          fclose(tempfp); fclose(htmlfp);
          remove(htmlnm);          /* delete old html file */
          rename(tmpnm, htmlnm);  /* rename temp to html name */
        }
        else {
          fclose(tempfp); fclose(htmlfp);  /* give up on curling */
          remove(tmpnm);
        }
      }
      else {      /* html file has been curled previously */
        printf("\nRecurling %s\n", htmlnm);
        if (write_middle(htmlnm, tempfp, htmlfp) == 1) {
                            /* put non-link text in temp */
          fclose(tempfp); fclose(htmlfp);
          tempfp = fopen(tmpnm,"r");
          htmlfp = fopen(htmlnm,"w");
          if (curl_temp(tmpnm, tempfp, htmlfp, hdstylfp, ftstylfp, 
                                          ent, top) == 1) {
            fclose(tempfp); fclose(htmlfp);
            remove(tmpnm);      /* delete temp now that html file recurled */
          }
          else {
            fclose(tempfp); fclose(htmlfp);
            remove(htmlnm);    
            rename(tmpnm, htmlnm);    /* text in temp becomes new html */
          }
        }
        else {
          fclose(tempfp); fclose(htmlfp);      /* give up on recurling */
          remove(tmpnm);
        }
      }
    }
  }
  else
    printf("\nCannot read %s\n", htmlnm);
  fclose(hdstylfp);
  fclose(ftstylfp);
}


void write_style(enum stylekind sk, FILE *fp, FILE *sfp, Entry *et)
/* Writes a style file (either for a header or footer) into
   the HTML file being generated. When a link word is found
   in the style file, it is replaced by the actual link. */
/* The labels allow the header and footer to be easily stripped
   out of a HTML file that is being relinked. */
/* no_nm is for the GIF's name when there isn't a link */
{
  enum linkkind lk;
  int c;
  char word[LINKLEN];
  char gifnm[PATHLEN];     /* holds the gif's complete file name */
  char no_nm[NAMELEN];     /* holds the no_gif's name */

  if (sk == FOOTER)
    fprintf(fp, "<!-- start-of-footer -->\n");     /* label footer start */
  while ((c = getc(sfp)) != EOF) {
    gifnm[0] = '\0';
    if (c == '%') {
      c = get_link(sfp, word, &lk);
      if ((lk == NONE) || (lk == INDEX) || (lk == FINDEX))
        fprintf(fp, "%c%s", '%', word);     /* to print a % */
      else {
        if (c == '(')
          c = get_gifnm(sfp, et, gifnm, no_nm);
        write_link(fp, word, gifnm, no_nm, lk, et);
      }
      if (c == EOF)
        break;
    }
    putc(c, fp);
  }
  if (sk == HEADER)
    fprintf(fp, "<!-- end-of-header -->\n");       /* label header end */
}


int get_link(FILE *fp, char word[], enum linkkind *plk)
/* The link words are detected. */
{
  int c, i;

  for (i=0; ((i < LINKLEN) &&
             ((c = getc(fp)) != EOF) && (isalpha(c))); i++)
    word[i] = c;
  word[i] = '\0';
  if (strcmp(word, "prev") == 0)
    *plk = PREV;
  else if (strcmp(word, "next") == 0)
    *plk = NEXT;
  else if (strcmp(word, "up") == 0)
    *plk = UP;
  else if (strcmp(word, "home") == 0)
    *plk = HOME;
  else if (strcmp(word, "cnts") == 0)
    *plk = INDEX;
  else if (strcmp(word, "fcnts") == 0)
    *plk = FINDEX;
  else if (strcmp(word, "keys") == 0)
    *plk = KEYS;
  else if (strcmp(word, "figs") == 0)
    *plk = FIGS;
  else if (strcmp(word, "tables") == 0)
    *plk = TBLS;
  else if (strcmp(word, "search") == 0)
    *plk = SRCH;
  else if (strcmp(word, "new") == 0)
    *plk = NEWPGS;
  else
    *plk = NONE;
  return c;
}


int get_gifnm(FILE *sfp, Entry *et, char gifnm[], char no_nm[])
/* Build the gif name, if it exists. */
/* no_nm is not preceded by an extension */
{
  int c, i;
  char name[NAMELEN];

  for (i=0; ((i < NAMELEN-3) && ((c = getc(sfp)) != EOF) && 
             (c != ')') && (c != '\n')); i++)
    name[i] = c;
  name[i] = '\0';
  if (c == ')')
    c = getc(sfp);      /* want character after the ')' */

  sprintf(no_nm, "no_%s", name);

  if (check_gifnm(name) == 1) {
    path_to_gifs(et->home->wd, gifnm);
    strcat(gifnm, name);
  }

  return c;
}


int check_gifnm(char *name)
/* Does the file arrows/name exist? */
{
  char chkname[PATHLEN];
  FILE *fp;

  strcpy(chkname, ARROWS);
  strcat(chkname, name);
  if((fp = fopen(chkname,"r")) == NULL) {
    printf("Style graphics file %s does not exist\n", chkname);
    return 0;
  }
  fclose(fp);
  return 1;
}
     

void path_to_gifs(char *home, char path[])
/* The path to the gifs is a combination of
   the path home and the directory arrows.
*/
{
  int posn, i;

  posn = strlen(home) - 1;   /* ignore '\0' entry */
  while ((posn >= 0) && (home[posn] != '/'))
    posn--;
  if (posn >= 0) {
    for(i=0; i <= posn; i++)
      path[i] = home[i];
    path[i] = '\0';
    strcat(path, ARROWS);
  }
  else
    strcpy(path, ARROWS);
}


void write_link(FILE *fp, char *word, char *gifnm, char *no_nm,
                              enum linkkind lk, Entry *et)
/* Depending on the link word in the style file, a different
   real URL link is written into the HTML file. */
/* Adonis El Fakih (eazl@gpd.plh.af.mil) suggested
   replacing No <word> by a suitable 'no_link' icon. */
{
  char no_gifnm[PATHLEN];
  char linknm[PATHLEN];

  switch (lk) {
    case PREV: strcpy(linknm, et->prev->wd); break;
    case NEXT: strcpy(linknm, et->next->wd); break;
    case UP  : strcpy(linknm, et->up->wd); break;
    case HOME: strcpy(linknm, et->home->wd); break;
    case KEYS: strcpy(linknm, et->key->wd); break;
    case FIGS: strcpy(linknm, et->fig->wd); break;
    case TBLS: strcpy(linknm, et->tbl->wd); break;
    case SRCH: strcpy(linknm, et->srch->wd); break;
    case NEWPGS: strcpy(linknm, et->newpgs->wd); break;
    case INDEX: case FINDEX: case NONE:
          printf("%c%s is an illegal header/footer link command\n", 
                        '%', word); 
          strcpy(linknm, ""); break;
  }
  word[0] = toupper(word[0]);     /* capitalise word */
  if (strcmp(linknm,"") == 0) {
    if (check_gifnm(no_nm) == 1) {   /* Based on Adonis' suggestion */
      path_to_gifs(et->home->wd, no_gifnm);
      strcat(no_gifnm, no_nm);
      fprintf(fp, "<img src=\"%s\" alt=\"No %s\">", no_gifnm, word);
    }
    else
      fprintf(fp, "No %s", word);
  }
  else if (gifnm[0] != '\0')
    fprintf(fp, "<a href=\"%s\"><img src=\"%s\" alt=\"%s\"></a>",
                  linknm, gifnm, word);
  else
    fprintf(fp, "<a href=\"%s\">%s</a>", linknm, word);
  
}


void write_cnt(FILE *hfp, Entry *et, Entry *top)
/* Write out an ordinary (single level) contents. */
/* Read the contents style file and use it as a template
   for the URL links to the Siblings of this entry. */
/* The labels allow the contents to be easily stripped out
   if the HTML file is being relinked. */
{
  int i;
  FILE *ifp;
  Sibs *kid;
  char line[MAXLINE], linestart[MAXLINE], lineend[MAXLINE];
  Entry *kidet;

  if (et->nokids == 0) {
    printf("Error: cannot have a contents in %s%s.html, since it has no children\n",
                          et->path->wd, et->name->wd);
    fprintf(hfp, "%ccnts\n", '%');
    return;
  }

  if ((ifp = fopen(et->indx->wd,"r")) != NULL) {
    fprintf(hfp, "<!-- start-of-cnts -->\n");    /* start label */
    while (fgets(line, MAXLINE, ifp) != NULL)
      if (linkline(line, linestart, lineend)) {
        kid = et->kids;
        for (i=0; i < et->nokids; i++) {
          kidet = find_entry(kid, top);    /* may return NULL */
          write_cntsline(hfp, et->home->wd, kid->fname->wd, kid, kidet, 
                                         linestart, lineend);
          kid = kid->nextsib;
        }
      }
      else
        fprintf(hfp, "%s", line);
    fprintf(hfp, "<!-- end-of-cnts -->\n");      /* end label */
    fclose(ifp);
  }
  else {
    printf("Cannot read %s contents style file, so contents not added to %s%s\n",
                 et->indx->wd, et->path->wd, et->name->wd);
    fprintf(hfp, "%ccnts\n", '%');
  }
}


void write_fcnts(FILE *hfp, Entry *et, Entry *top)
/* Write out an all-levels contents. */
/* Read the contents style file and use it as a template
   for the links to all the child entries below this entry. */
/* The Sibs d.s. is no longer sufficient; must search
   the Entries d.s. */
/* The labels allow the contents to be easily stripped out
   if the HTML file is being relinked. */
{
  int i;
  FILE *ifp;
  Sibs *kid;
  char line[MAXLINE], linestart[MAXLINE], lineend[MAXLINE];

  if (et->nokids == 0) {
    printf("Error: cannot have a full contents in %s%s.html, since it has no children\n",
                          et->path->wd, et->name->wd);
    fprintf(hfp, "%cfcnts\n", '%');
    return;
  }

  if ((ifp = fopen(et->indx->wd,"r")) != NULL) {
    fprintf(hfp, "<!-- start-of-fcnts -->\n");    /* start label */
    while (fgets(line, MAXLINE, ifp) != NULL)
      if (linkline(line, linestart, lineend)) {
        kid = et->kids;
        for (i=0; i < et->nokids; i++) {
          next_cnt(hfp, et->home->wd, kid, "", top, linestart, lineend);
          kid = kid->nextsib;
        }
      }
      else
        fprintf(hfp, "%s", line);
    fprintf(hfp, "<!-- end-of-fcnts -->\n");      /* end label */
    fclose(ifp);
  }
  else {
    printf("Cannot read %s contents style file, so full contents not added to %s%s\n",
                    et->indx->wd, et->path->wd, et->name->wd);

    fprintf(hfp, "%cfcnts\n", '%');
  }
}


int linkline(char line[], char lstart[], char lend[])
/* Split a contents style file line which contains %item into 2 parts:
   the part of the line before, and the part after. The
   place where %item was will be replaced by the real contents
   URL link. */
{
  int i, s, linkflag;

  s = 0; linkflag = 0;
  for (i=0; ((line[i] != '\0') && (!linkflag)); i++) {
     if (same_chars(line,i,"%item") == 1) {
       lstart[i] = '\0';
       strcpy(lend, &(line[i+5]));
       linkflag = 1;
     }
     else
       lstart[i] = line[i];
  }
  return linkflag;
}


void write_cntsline(FILE *fp, char *pathhome, char *pathfnm, 
                       Sibs *kid, Entry *et, char *lnstart, char *lnend)
/* Write out the contents line with the real link. 
   Obtain the TITLE string from the file; if none
   there use the kid's name. */
{
  char string[MAXLINE];
  char gifnm[PATHLEN];   /* path to NEW gif */

  if (et != NULL)
    read_title(et, kid->name->wd, string);
  else
    strcpy(string, kid->name->wd);
  fprintf(fp, "%s", lnstart);
  fprintf(fp, "<a href=\"%s\">%s</a>", pathfnm, string);
  if (et->new == 1) {
    path_to_gifs(pathhome, gifnm);
    strcat(gifnm, NEWGIF);
    fprintf(fp, "  <img src=\"%s\" alt=\"NEW!\">", gifnm);
  }
  fprintf(fp, "%s", lnend);
}


void read_title(Entry *et, char *name, char string[])
/* The <TITLE> line is looked for in the first READLNS lines of
   a file. The format is:
   <TITLE>string-bits</TITLE>
   If there isn't a TITLE line then the name is used as the string.
*/
{
  FILE *fp;
  char line[MAXLINE];
  int i, titlefound;
  char ffn[PATHLEN];

  sprintf(ffn, "%s%s.html", et->path->wd, et->name->wd);
  if ((fp = fopen(ffn,"r")) == NULL) {
    printf("Cannot open %s to read TITLE line\n", ffn); /* shouldn't happen */
    strcpy(string, name);
  }
  else {
    titlefound = 0;
    for (i=0; ((i < READLNS) && (fgets(line, MAXLINE, fp) != NULL)
                             && (!titlefound)); i++)
      if (find_title(ffn, line, string))
        titlefound = 1;
    fclose(fp);
    if (titlefound)
      printf("<TITLE> line found in %s\n", ffn);
    else
      strcpy(string, name);
  }
}


int find_title(char *ffn, char line[], char string[])
/* The <TITLE> line in a file has the form:
      <TITLE>string-bit</TITLE>
   or <title>string-bit</title>
*/
{
  int i, k;

  if (start_title(line, 0) == 1) {
    i = 7;
    k = 0;
    while ((end_title(line, i) == 0) && (i < MAXLINE))
      string[k++] = line[i++];
    string[k] = '\0';
    if (i == MAXLINE) {
      printf("In %s, Could not find </TITLE> in line:   %s", ffn, line);
      return 0;
    }
    else
      return 1;
  }
  return 0;
}


int start_title(char line[], int i)
/* check for <TITLE> or <title> */
{
  return (same_chars(line, i, "<TITLE>") ||
          same_chars(line, i, "<title>"));
}


int end_title(char line[], int i)
/* check for </TITLE> or </title> */
{
  return (same_chars(line, i, "</TITLE>") ||
          same_chars(line, i, "</title>"));
}



void next_cnt(FILE *fp, char *pathhome, Sibs *kdy, char *path, 
                   Entry *top, char *lnstart, char *lnend)
/* Generate the all-level contents. For a given
   entry output all of its children, which are processed
   by recursive calls to next_cnt so that their children
   can be processed. */
{
  char *npath, *pathfnm;
  Entry *et;
  Sibs *kid;
  int i;

  if ((et = find_entry(kdy, top)) != NULL) {
    npath = extend_path(path, et->dir->wd);
    pathfnm = add_path_html(npath, kdy->name->wd);
    write_cntsline(fp, pathhome, pathfnm, kdy, et, lnstart, lnend);
    if (et->nokids > 0) {
      kid = et->kids;
      fprintf(fp, "<ul>\n");
      for (i=0; i < et->nokids; i++) {
        next_cnt(fp, pathhome, kid, npath, top, lnstart, lnend);
        kid = kid->nextsib;
      }
      fprintf(fp, "</ul> <p>\n");
    }
    free(npath);
    free(pathfnm);
  }
}


Entry *find_entry(Sibs *kid, Entry *top)
{
  char ns[PATHLEN];
  Entry *et;

  et = top;
  while (et != NULL) {
    if (strcmp(et->name->wd, kid->name->wd) == 0) {
      if (et->dir->wd[0] == '\0')       /* no directory above */
        sprintf(ns, "%s.html", et->name->wd);
      else
        sprintf(ns, "%s/%s.html", et->dir->wd, et->name->wd);
      if (strcmp(ns, kid->fname->wd) == 0)
        return et;
    }
    et = et->nextentry;
  }
  printf("Unable to find a contents entry for %s\n",
                     kid->name->wd);
  return NULL;
}


int last_header(FILE *fp)
/* Scan the entire file, and record the last line
   which has a "<!-- end-of-header -->". */
{
  char line[MAXLINE];
  int lineno, last;

  lineno = 0;
  last = -1;
  while (fgets(line, MAXLINE, fp) != NULL) {
    lineno++;
    if (strcmp(line, "<!-- end-of-header -->\n") == 0)
      last = lineno;
  }
  rewind(fp);
  return last;
}


int curl_temp(char *htmlnm, FILE *fromfp, FILE *tofp, FILE *hdsfp,
                FILE *ftsfp, Entry *ent, Entry *top)
/* Using the file fromfp which contains html text, create tofp
   which includes the curl links .
*/
{
  int i;
  int bodystart = -1;
  int bodyend = -1;
  char line[MAXLINE];

  find_body(htmlnm, fromfp, &bodystart, &bodyend);
  if (bodystart == -1) {
    printf("%s being ignored since it has no <body> tag\n", htmlnm);
    return 0;
  }
  if (bodyend == -1) {
    printf("%s being ignored since it has no </body> tag\n", htmlnm);
    return 0;
  }
  for (i=1; i <= bodystart; i++) {    /* skip to start of body */
    fgets(line, MAXLINE, fromfp);
    fputs(line, tofp);
  }
  write_style(HEADER, tofp, hdsfp, ent);    /* write header */
  for (i = (bodystart+1); i < bodyend; i++) {
    fgets(line, MAXLINE, fromfp);
    if (strcmp(line, "%cnts\n") == 0)
      write_cnt(tofp, ent, top);
    else if (strcmp(line, "%fcnts\n") == 0)
      write_fcnts(tofp, ent, top);
    else
      fputs(line, tofp);
  }
  write_style(FOOTER, tofp, ftsfp, ent);     /* write footer */
  while (fgets(line, MAXLINE, fromfp) != NULL)  /* output rest */
    fputs(line, tofp);
  return 1;
}


void find_body(char *nm, FILE *fp, int *start, int *end)
/* Set start to be the line number containing <body ... 
   (don't look for '>' since there may be atttributes);
   end is assigned the line number containing </body>.
   They are -1 by default.
   The tags need to start the line.
*/
{
  char line[MAXLINE];
  int lineno = 0;

  while (fgets(line, MAXLINE, fp) != NULL) {
    lineno++;
    if ((same_chars(line, 0, "<body") == 1) ||
        (same_chars(line, 0, "<BODY") == 1)) {
      *start = lineno;
      break;
    }
  }
  if (*start != -1)
    while (fgets(line, MAXLINE, fp) != NULL) {
      lineno++;
      if ((same_chars(line, 0, "</body>") == 1) ||
          (same_chars(line, 0, "</BODY>") == 1)) {
      *end = lineno;
      break;
      }
    }
  rewind(fp);
}


int write_middle(char *nm, FILE *tfp, FILE *hfp)
/* Extracts the textual 'middle' of a HTML file.
   This means stripping off the header and footer parts.
   Also any cnts/fcnts entries are replaced by %cnts/%fcnts labels
   so that they can be reprocessed later. 
   The links attached to Figures and Tables are also removed.
*/
{
  char line[MAXLINE], nline[MAXLINE];
  int i = 0;
  int footflag = 0;
  int bodystart = -1;
  int bodyend = -1;

  find_body(nm, hfp, &bodystart, &bodyend);
  if (bodystart == -1) {
    printf("%s being ignored since it has no <body> tag\n", nm);
    return -1;
  }
  if (bodyend == -1) {
    printf("%s being ignored since it has no </body> tag\n", nm);
    return -1;
  }
  while (i < bodystart) {    /* skip to start of body */
    fgets(line, MAXLINE, hfp);
    i++;
    fputs(line, tfp);
  }

  while (strcmp(line, "<!-- end-of-header -->\n") != 0) {
    fgets(line, MAXLINE, hfp);
    i++;
  }

  while (i < bodyend) {
    fgets(line, MAXLINE, hfp);
    i++;
    if (strcmp(line, "<!-- start-of-cnts -->\n") == 0) {
      while (strcmp(line, "<!-- end-of-cnts -->\n") != 0)
        if (fgets(line, MAXLINE, hfp) == NULL) {
          printf("Missing <!-- end-of-cnts -->\n");
          return -1;
        }
        else i++;
      fprintf(tfp, "%ccnts\n", '%');          /* want %cnts printed */
    }
    else if (strcmp(line, "<!-- start-of-fcnts -->\n") == 0) {
      while (strcmp(line, "<!-- end-of-fcnts -->\n") != 0)
        if (fgets(line, MAXLINE, hfp) == NULL) {
          printf("Missing <!-- end-of-fcnts -->\n");
          return -1;
        }
        else i++;
      fprintf(tfp, "%cfcnts\n", '%');         /* want %fcnts printed */
    }
    else if (same_chars(line, 0, "<b>Figure</b>:") == 1) {
      strip_pic(line, nline);
      fputs(nline, tfp);
    }
    else if (same_chars(line, 0, "<b>Table</b>:") == 1) {
      strip_pic(line, nline);
      fputs(nline, tfp);
    }
    else if (strcmp(line, "<!-- start-of-footer -->\n") == 0) {
      footflag = 1;
      while (i < bodyend) {
        fgets(line, MAXLINE, hfp);
        i++;
      }
      fputs(line, tfp);
    }
    else
      fputs(line, tfp);
  }
  if (!footflag)      /* no footer encountered */
    printf("  Footer information missing\n");
  while (fgets(line, MAXLINE, hfp) != NULL)
    fputs(line, tfp);
  return 1;
}


int same_chars(char line[], int index, char pat[])
/* Is pat in line starting at index? */
{
  int i;

  for (i=0; pat[i] != '\0'; i++)
    if (line[index+i] != pat[i])
      return 0;
   return 1;
}


void strip_pic(char line[], char nline[])
/* Build a new line where the pic information after the
   first '.' has been deleted, and replaced by <p>. */
{
  int k = 0;

  while (line[k] != '.') {
    nline[k] = line[k];
    k++;
  }
  nline[k] = '\0';
  strcat(nline, ". <p>\n");
}



void save_entries(Entry *new)
/* Save the Entry d.s. into ENTS */
{
  Entry *e;
  Sibs *ks;
  FILE *efp;

  if ((efp = fopen(ENTS, "w")) == NULL)
    printf("Could not open %s for writing\n", ENTS);
  else {
    e = new;
    while (e != NULL) {
      fprintf(efp, "%s %s %s %d %s %s %s %s %s %s %s %s %s %s %s %s %d ",
           star(e->name->wd), star(e->path->wd), star(e->dir->wd), e->new,
           star(e->prev->wd), star(e->next->wd), star(e->up->wd),
           star(e->home->wd), star(e->key->wd), star(e->fig->wd),
           star(e->tbl->wd), star(e->srch->wd), star(e->newpgs->wd),
           star(e->header->wd), star(e->footer->wd), 
           star(e->indx->wd), e->nokids);
      if (e->nokids > 0) {
        ks = e->kids;
        while (ks != NULL) {
          fprintf(efp, "%s ", ks->name->wd);
          fprintf(efp, "%s ", ks->fname->wd);
          ks = ks->nextsib;
        }
      }
      fputc('\n', efp);
      e = e->nextentry;
    }
    fclose(efp);
  }
}

char *star(char *s)
/* Save empty strings as *'s */
{
  if (strcmp(s,"") == 0)
    return "*";
  else
    return s;
}


void announce_old(Entry *old)
/* Announce names (and paths) which are no longer referenced
   by current contents. Free Entry d.s. during announcement. */
{
  Entry *et;
  char *ffn;

  if (old != NULL) {
    printf("\nThe following names were in the previous version of the contents file, but not this one:\n");
    while (old != NULL) {
      ffn= add_path(old->path->wd, old->name->wd);
      printf("  %s\n", ffn);
      free(ffn);
      et = old;
      old = old->nextentry;
      free_entry(et);
      et = NULL;
    }
  }
}


/* Create/update the keys file */

void collect_keys(char *htmlnm, Entry *ent, Key **pkeys)
/* For a given file, collect its keys into ekeys, and
   then add them to the linked list pointed to by pkeys. */
{
  int keyno, i;
  Entkey ekeys[EKEYNO];

  if (read_ekeys(htmlnm, ekeys, &keyno))
    for (i=0; i < keyno; i++)
      add_to_keys(ekeys[i], ent->name->wd, htmlnm, pkeys);
}


int read_ekeys(char *ffn, Entkey ekeys[], int *pkeyno)
/* The keyword line is looked for in the first READLNS lines of
   a file. */
{
  FILE *fp;
  char line[MAXLINE];
  int i, keyfound;

  if ((fp = fopen(ffn,"r")) == NULL) {
    printf("Cannot open %s to read keywords\n", ffn);
    return 0;
  }
  else {
    *pkeyno = 0;
    keyfound = 0;
    for (i=0; ((i < READLNS) && (fgets(line, MAXLINE, fp) != NULL) 
                             && (!keyfound)); i++)
      if (find_ekey(line, ekeys, pkeyno))
        keyfound = 1;
    fclose(fp);
    if (keyfound)
      printf("Keys found in %s\n", ffn);
    return keyfound;
  }
}


void add_to_keys(Entkey ek, char *nm, char *ffn, Key **pkeys)
/* Add a keyword from the nm entry (file name ffn) into
   the keys list. The list is in ascending order by keyword,
   and the nm entries (the refs) are in ascending order for
   each keyword. */
{
  Key *curr, *prev;

  if (*pkeys == NULL)
    *pkeys = new_key(ek, nm, ffn);
  else {
    curr = *pkeys;
    prev = NULL;
    while ((curr != NULL) && (strcmp(curr->word,ek.word) < 0)) {
      prev = curr;
      curr = curr->nextkey;
    }
    if (curr == NULL)
      prev->nextkey = new_key(ek, nm, ffn);
    else if (strcmp(curr->word,ek.word) > 0) {
      if (prev == NULL) {    /* first el is bigger */
        *pkeys = new_key(ek, nm, ffn);
        (*pkeys)->nextkey = curr;
      }
      else {
        prev->nextkey = new_key(ek, nm, ffn);
        prev->nextkey->nextkey = curr;
      }
    }
    else    /* curr->word == ek.word */
      add_ref( &(curr->krs), ek, nm, ffn);
  }
}


Key *new_key(Entkey ek, char *nm, char *ffn)
/* A new key is composed from the key word in ek, and
   the nm/ffn details which becomes the ref in the
   KEYSHTML file. */
{
  Key *temp;
  temp = (Key *) malloc(sizeof(Key));
  temp->word = sassign(ek.word);
  temp->krs = new_ref(nm, ffn, ek.rk);
  temp->nextkey = NULL;
  return temp;
}


Keyref *new_ref(char *nm, char *ffn, enum refkind rk)
/* A ref is a nm/ffn pair which is used to
   generate the URL link to the file ffn labelled by nm 
   in KEYSHTML */
{
  Keyref *temp;
  temp = (Keyref *) malloc(sizeof(Keyref));
  temp->nm = sassign(nm);
  temp->fullfn = sassign(ffn);
  temp->rk = rk;
  temp->nextref = NULL;
  return temp;
}


void add_ref(Keyref **krs, Entkey ek, char *nm, char *ffn)
/* A keyword points to a list of refs in alphabetical
   order. Add a new reference to this d.s. */
{
  Keyref *curr, *prev;

  curr = *krs;
  prev = NULL;
  while ((curr != NULL) && (strcmp(curr->nm,nm) < 0)) {
    prev = curr;
    curr = curr->nextref;
  }
  if (curr == NULL)
    prev->nextref = new_ref(nm, ffn, ek.rk);
  else if (strcmp(curr->nm, nm) > 0) {
    if (prev == NULL) {    /* first el is bigger */
      *krs = new_ref(nm, ffn, ek.rk);
      (*krs)->nextref = curr;
    }
    else {
      prev->nextref = new_ref(nm, ffn, ek.rk);
      prev->nextref->nextref = curr;
    }
  }
  else    /* curr->nm == nm */
    printf("Two uses of the keyword %s by %s; second being ignored\n",
                ek.word, nm);
}


int find_ekey(char line[], Entkey keys[], int *keyno)
/* The keywords line in a file has the form:
      <b>Keywords</b>: k1, k2, -k3, k4. <p>
   The '-' signals that k3 should be removed from KEYSHTML.
   The '.' is important.
   The keys array is filled with the words k1, k2, etc and
   whether they are being added (the default) or removed
   from KEYSHTML. 
   The words are converted into lowercase. */
{
  int i, k;
  char word[NAMELEN];

  if (same_chars(line, 0, "<b>Keywords</b>:") == 1) {
    for (i=16; line[i] == ' '; i++);
    if (line[i] == '\0') {
      printf("Syntax error in keywords line:   %s", line);
      return 0;
    }
    *keyno = 0;
    while ((line[i] != '.') && (line[i] != '\0')) {  /* terminating symbol */
      k = 0;
      if (line[i] == '-') {
        keys[*keyno].rk = DEL;
        i++;
      }
      else
        keys[*keyno].rk = ADD;
      while ((line[i] != ',') && (line[i] != '.') && (line[i] != '\0'))
        word[k++] = line[i++];
      word[k] = '\0';
      make_lower(word);
      if (line[i] == '\0') {
        printf("Keywords line has no '.':   %s", line);
        return 0;
      }
      if (islower(word[0])) {
        keys[*keyno].word = sassign(word);
        (*keyno)++;
      }
      else 
        printf("Keyword %s does not begin with a letter; being ignored\n",
                               word);
      if (line[i] == ',') {
        i++;
        while (line[i] == ' ') i++;
      }
    }
    if (line[i] == '\0') {
      printf("Syntax error in keywords line:   %s", line);
      return 0;
    }
    else
      return 1;
  }
  return 0;
}

void make_lower(char word[])
/* Convert the word to lowercase. */
{
  int i = 0;
  while (word[i] != '\0') {
    word[i] = tolower(word[i]);
    i++;
  }
}


/* Some print utilities for debugging Key d.s. */

void print_keys(Key *ks)
{
  Key *k;
  k = ks;
  while (k != NULL) {
    printf(" word is: %s\n", k->word);
    printf(" keyref is:\n");
    print_keyref(k->krs);
    k = k->nextkey;
  }
  printf("\n\n");
}

void print_keyref(Keyref *krs)
{
  Keyref *kr;
  kr = krs;
  while (kr != NULL) {
    printf(" nm is: %s\n", kr->nm);
    printf(" fullfn is: %s\n", kr->fullfn);
    printf(" rk is: "); print_refkind(kr->rk); putchar('\n');
    putchar('\n');
    kr = kr->nextref;
  }
}

void print_refkind(enum refkind rk)
{
  switch(rk) {
    case ADD: printf("ADD"); break;
    case DEL: printf("DEL"); break;
  }
}


void update_kfile(Key *keys)
/* Modify KEYSHTML by writing a new version to a temporary
   key file. The old KEYSHTML is then overwritten. */
{
  Key *temp;
  char line[MAXLINE], tmpnm[PATHLEN];
  FILE *kfp, *tkfp;
  int c, prevch;

  if (keys != NULL) {
    printf("\nKeys Processing:\n");
 /* printf("\nkeys are:\n");
    print_keys(keys); */

    if ((kfp = fopen(KEYSHTML,"r")) == NULL) {
      printf("Cannot open %s\n", KEYSHTML);
    }
    else {
      strcpy(tmpnm, KEYSHTML);
      strcat(tmpnm, "tmp");
      if ((tkfp = fopen(tmpnm,"w")) == NULL) {
        printf("Cannnot open temporary key file\n");
        fclose(kfp);
      }
      else {
        prevch = ' ';
        fgets(line, MAXLINE, kfp);
        while (keys != NULL) {
          modify_kf(keys, kfp, tkfp, line, prevch);
          prevch = keys->word[0];
          temp = keys;
          keys = keys->nextkey;
          free_key(temp);
        }
        fputs(line, tkfp);
        while ((c = getc(kfp)) != EOF)
          putc(c, tkfp);
        fclose(kfp); fclose(tkfp);
        remove(KEYSHTML);
        rename(tmpnm, KEYSHTML);
      }
    }
  }
}


void modify_kf(Key *keys, FILE *kfp, FILE *tkfp, char *line, int prevch)
/* Copy a line at a time until the beginning of the key words
   which start with the first letter of the key being inserted.
   Insert the key in alphabetical order with the other key words. */
{
  int ch, kfound;

  ch = keys->word[0];
  copy_to_entries(kfp, tkfp, prevch, line, ch);   /* get to start of keys */
  while((kfound = find_key(line, keys->word)) < 0) {
    fputs(line, tkfp);
    fgets(line, MAXLINE, kfp);
  }
  if (kfound == 0) {          /* line key == keyword */
    hrefs_inserts(line, keys);   /* insert ref into existing key line */
    fputs(line, tkfp);
    fgets(line, MAXLINE, kfp);
  }
  else     /* kfound > 0 */     /* line key > keyword or </ul> */
    key_inserts(keys, tkfp);    /* create new key word line */
}


void hrefs_inserts(char *line, Key *keys)
/* The key pointed to by Keys must be inserted into an
   existing key word line. Deletion is also done here. 
   The format of a key word line is:
     <li> word : <a href="..">nm1</a>, <a href="..">nm2</a>, .... <p> 
*/
{
  char kword[NAMELEN];
  int i, cmp;
  Keyref *refs;

  strcpy(kword, keys->word);
  refs = keys->krs;
  i = 0;
  while ((same_chars(line, i, "<p>") == 0) && (refs != NULL)) {
    if ((line[i] == '<') && (line[i+1] == 'a')) {
      if ((cmp = html_cmp(line, i, refs->nm)) == 0) {   /* line ref == ref */
        if (refs->rk == DEL) {
          printf("%s ref being deleted from key %s\n",
                               refs->nm, kword);
          delete_lineref(line, i);
        }
        else if ((cmp = ffn_cmp(line, i, refs->fullfn)) != 0) { 
          /* line ref == ref but ffn's different */
          printf("Key %s with %s ref in old position being updated\n", 
                                                     kword, refs->nm);
          delete_lineref(line, i);
          insert_ref(refs, line, &i, kword);
        }
        else
          printf("%s key already has a %s reference\n", kword, refs->nm);
        refs = refs->nextref;
      }
      else if (cmp > 0) {         /* line ref > ref */
        insert_ref(refs, line, &i, kword);
        refs = refs->nextref;
      }
      /* else cmp < 0 */
    }
    i++;       /* carry on along line ref */
  }
  if (same_chars(line, i, "<p>") == 1) 
    while (refs != NULL) {
      insert_ref(refs, line, &i, kword);
      refs = refs->nextref;
    }
}


void delete_lineref(char *line, int i)
/* Delete a reference by deleting all the characters 
   in the line between the ith position and <a or <p. 
   This makes sense when you look at the format of 
   a line. */
{
  int fin, x;

  fin = i+1;
  while (a_or_p(line,fin) == 0)
    fin++;
  for (x=0; line[fin+x] != '\0'; x++)
    line[i+x] = line[fin+x];
  line[i+x] = '\0';
}


void insert_ref(Keyref *refs, char *line, int *pi, char *kword)
/* Insert a single reference into a key word line. */
{
  if (refs->rk == DEL)
    printf("%s reference cannot be deleted since not in %s key\n",
            refs->nm, kword);
  else {
    printf("%s ref being added to key %s\n", refs->nm, kword);
    add_string(line, pi, "<a href=\"");
    add_string(line, pi, refs->fullfn);
    add_string(line, pi, "\">");
    add_string(line, pi, refs->nm);
    add_string(line, pi, "</a>   ");
  }
}


void add_string(char *line, int *pi, char *s)
/* Insert the string s into line beginning at *pi.
   Adjust *pi to be pointing to just after s. */
{
  int i, slen, linelen;

  slen = strlen(s);
  linelen = strlen(line)+1;
  if (linelen+slen >= MAXLINE) {
    printf("Line too long: cannot add \"%s\" string to the line \"%s\"\n",
                    s, line);
  }
  else {
    for (i=linelen; i >= (*pi); i--)
      line[i+slen] = line[i];
    for (i=0; i < slen; i++)
      line[(*pi)+i] = s[i];
    *pi = (*pi) + slen;
  }
}


void key_inserts(Key *keys, FILE *tkfp)
/* Create a new key word line. */
{
  Keyref *refs;
  char line[MAXLINE];
  int i = 0;

  add_string(line, &i, "<li> ");
  add_string(line, &i, keys->word);
  add_string(line, &i, ": ");
  refs = keys->krs;
  while (refs != NULL) {
    insert_ref(refs, line, &i, keys->word);
    refs = refs->nextref;
  }
  add_string(line, &i, "<p>\n");
  line[i] = '\0';
  fputs(line, tkfp);
}


void copy_to_entries(FILE *kfp, FILE *tkfp, int prevch, char *line, int ch)
/* Write out a line at a time until a comment line for ch 
   is reached. Write out the next two lines, and return with the third.
   This is due to the format of KEYSHTML for each alphabetical entry:
     <!-- ch -->
     HTML code for a ch title
     <ul>
     First keyword line for ch (or </ul>)

   If ch is the same as last time, do nothing since already in
   the midst of the right keys.
*/
   
{
  if (prevch != ch) {      /* ch is a new letter */
    while (char_comment(line, ch) == 0) {     /* get to comment line for ch */
      fputs(line, tkfp);
      fgets(line, MAXLINE, kfp);
    }
    fputs(line, tkfp);
    fgets(line, MAXLINE, kfp);    /* the HTML line for ch */
    fputs(line, tkfp);
    fgets(line, MAXLINE, kfp);    /* the HTML <ul> line */
    fputs(line, tkfp);
    fgets(line, MAXLINE, kfp);    /* first <li> line for ch or </ul> */
  }
}

int char_comment(char *line, int ch)
/* Does the comment line start with <!-- ch ? */
{
  if ((line[0] == '<') && (line[1] == '!') &&
      (line[2] == '-') && (line[3] == '-' ) &&
      (line[4] == ' ') && (line[5] == ch ))
    return 1;
  else
    return 0;
}


int find_key(char *line, char *key)
/* Is key the same as the key word (or words) on this line? 
   The format is:
      <li> words : .....
*/
{
  int i, w;
  char word[NAMELEN];

  if (same_chars(line,0,"<li>") == 1) {
    for (i=4; line[i] == ' '; i++);
    w = 0;
    while (line[i] != ':')
      word[w++] = line[i++];
    word[w] = '\0';
    return strcmp(word, key);
  }
  return 1;
}


int html_cmp(char *line, int i, char *nm)
/* Get the name of the ref, and compare it with nm.
   The i starts at <a in the line below.
   <a href="..">name</a>,
*/
{
  char hword[NAMELEN];
  int w = 0;

  while (same_chars(line, i, "\">") == 0)
    i++;
  i = i+2;
  while (line[i] != '<')
    hword[w++] = line[i++];
  hword[w] = '\0';
  return strcmp(hword, nm);
}


int ffn_cmp(char *line, int i, char *refffn)
/* Get the full filename of the ref, and compare it with refffn.
   The i starts at <a in the line below.
   <a href="..">name</a>,
*/
{
  char ffn[PATHLEN];
  int w = 0;

  while (line[i] != '\"')
    i++;
  i++;
  while (line[i] != '\"')
    ffn[w++] = line[i++];
  ffn[w] = '\0';
  return strcmp(ffn, refffn);
}


int a_or_p(char *line, int i)
/* Return 1 if <a or <p has been reached in the line, 0 otherwise */
{
  if ((line[i] == '<') &&
      ((line[i+1] == 'a') || (line[i+1] == 'p')) )
    return 1;
  else
    return 0;
}


/* Reclaim Key d.s. */

void free_key(Key *k)
{
  if (k != NULL) {
    free(k->word);
    free_krefs(k->krs);
    free(k);
  }
}

void free_krefs(Keyref *kr)
{
  Keyref *temp;

  while (kr != NULL) {
    free(kr->nm);
    free(kr->fullfn);
    temp = kr;
    kr = kr->nextref;
    free(temp);
  }
}


/* collect Pic information */

void collect_pics(char *ffn, Entry *ent, Pic **pfigs, Pic **ptbls)
/* Figure and table lines are gathered from the file,
   and the figure/table line is modified to have a link
   to the keys.html/tbls.html file. 
*/
{
  FILE *fp, *tempfp;
  char line[MAXLINE], tmpnm[PATHLEN];
  int lineno = 1;

  if ((fp = fopen(ffn,"r")) == NULL)
    printf("Cannot open %s to read figures and tables\n", ffn);
  else {
    strcpy(tmpnm, ffn);
    strcat(tmpnm, "tmp");
    if ((tempfp = fopen(tmpnm,"w")) == NULL) {
      printf("\nCannot open temporary file for %s\n", ffn);
      remove(tmpnm);
      fclose(fp);
    }
    else {
      while (fgets(line, MAXLINE, fp) != NULL) {
        if (find_pics(ffn, line, lineno, "<b>Figure</b>:", "Figure", 
                                ent, tempfp, pfigs) == 1)
          printf("Modified Figure line: %s", line);
        else if (find_pics(ffn, line, lineno, "<b>Table</b>:", "Table", 
                                ent, tempfp, ptbls) == 1)
          printf("Modified Table line: %s", line);
        else
          fputs(line, tempfp);
        lineno++;
      }
      fclose(fp); fclose(tempfp);
      remove(ffn);
      rename(tmpnm, ffn);
    }
  }
}


int find_pics(char *ffn, char line[], int lineno, char *piclabel,
          char *picword, Entry *ent, FILE *tfp, Pic **pp)
/* The pics line in a file has the form:
     piclabel string. rest-of-line
   piclabel is either '<b>Figure</b>:' or '<b>Table</b>:'
   The '.' is important. 
*/
/* It is modified to become:
     <b>picword</b>: string. <a href="ent->fig (or tbl)">To picword file</a>
                                   <a name="lineno">.</a> <p>
   The href goes to the FIGSHTML or TBLSHTML file. The name anchor
   allows this line to be reached from the FIGSHTML or TBLSHTML file.
*/
/* Store the pics information in the pp list, and output
   the modified line. */
{
  int i, k;
  char desc[MAXLINE], nline[MAXLINE];

  if (same_chars(line, 0, piclabel) == 1) {
    for (i=strlen(piclabel); line[i] == ' '; i++);
    if (line[i] == '\0') {
      printf("Syntax error in %s line:   %s", picword, line);
      return 0;
    }
    k = 0;
    while ((line[i] != '.') && (line[i] != '\0'))  /* terminating symbol */
      desc[k++] = line[i++];
    desc[k] = '\0';
    if (line[i] == '\0') {
      printf("%s line has no '.' :   %s", picword, line);
      return 0;
    }
    build_picline(line, i, ent, picword, lineno, nline);
    fputs(nline, tfp);
    
    add_to_pics(ffn, desc, lineno, pp);
    return 1;
  }
  else
    return 0;
}


void build_picline(char line[], int i, Entry *ent, char *picword, 
                                        int lineno, char nline[])
/* The picline is modified to become:
     picword: string. <a href="ent->fig->wd (or tbl->wd)">To picword s</a>
                                     <a name="lineno">.</a> <p>
*/
{  
  char nostring[NAMELEN];
  int k;

  for (k=0; k <= i; k++)     /* copy existing line up to '.' */
    nline[k] = line[k];
  add_string(nline, &k, " <a href=\"");
  if (strcmp(picword, "Figure") == 0)
    add_string(nline, &k, ent->fig->wd);
  else
    add_string(nline, &k, ent->tbl->wd);
  add_string(nline, &k, "\">To ");
  add_string(nline, &k, picword);
  add_string(nline, &k, "s</a><a name=\"");
  itoa(lineno, nostring);
  add_string(nline, &k, nostring);
  add_string(nline, &k, "\">. </a> <p>\n");
  nline[k] = '\0';
}


void itoa(int n, char s[])
/* Convert a (positive) integer into a string. */
/* See K&R p.64 */
{
  int i, j;
  char rev[NAMELEN];

  i = 0; 
  do {      /* generate digits in reverse order */
    rev[i++] = n % 10 + '0';
  } while ((n /= 10) > 0);

  j = 0;
  while (i > 0) {
    s[j] = rev[i-1];
    j++; i--;
  }
  s[j] = '\0'; 
}


void add_to_pics(char *ffn, char *desc, int lineno, Pic **pp)
/* Add a description of a figure/table from the file ffn to
   the end of the pp list. */
{
  Pic *last;

  if (*pp == NULL)
    *pp = new_pic(desc, ffn, lineno);
  else {
    last = *pp;
    while (last->nextpic != NULL)
      last = last->nextpic;
    last->nextpic = new_pic(desc, ffn, lineno);
  }
}


Pic *new_pic(char *desc, char *ffn, int lineno)
/* A new pic is made up of a description (of a figure or
   table) and its file location. */
{
  Pic *temp;
  temp = (Pic *) malloc(sizeof(Pic));
  temp->title = sassign(desc);
  temp->fullfn = sassign(ffn);
  temp->lineno = lineno;
  temp->nextpic = NULL;
  return temp;
}


/* Update Pic information in FIGSHTML/TBLSHTML */

void mod_picfile(Pic *pics, char *ftitle, char *picfn)
/* Generate a new pic file (either FIGSHTML or TBLSHTML)
   until a <ul> line is reached, then remove any lines
   referring to the titles or files in the pics d.s, and add
   the new lines at the end of the list. After the
   </ul> line output the file unchanged. */
/* A line has the form:
     <li><a href="pics->fullfn#pic->lineno">pics->title</a> <p>
*/
{
  FILE *fp, *tempfp;
  char line[MAXLINE], tmpnm[PATHLEN];
  int c;

  if (pics != NULL) {
    if ((fp = fopen(picfn,"r")) == NULL)
      printf("Cannot open %s\n", picfn);
    else {
      strcpy(tmpnm, picfn);
      strcat(tmpnm, "tmp");
      if ((tempfp = fopen(tmpnm,"w")) == NULL) {
        printf("\nCannot open temporary file for %s\n", picfn);
        remove(tmpnm);
        fclose(fp);
      }
      else {
        printf("Modifying %s\n", picfn);
        fgets(line, MAXLINE, fp);
        while (same_chars(line, 0, "<ul>") == 0) {
          fputs(line, tempfp);
          fgets(line, MAXLINE, fp);
        }
        fputs(line, tempfp);         /* skip over <ul> line */
        fgets(line, MAXLINE, fp);
        while (same_chars(line, 0, "</ul>") == 0) {
          delete_oldpics(pics, line, tempfp);
          fgets(line, MAXLINE, fp);
        }
        insert_newpics(pics, tempfp);
        fputs(line, tempfp);       /* write out the </ul> line */
        while ((c = getc(fp)) != EOF)
          putc(c, tempfp);

        fclose(fp); fclose(tempfp);
        remove(picfn);
        rename(tmpnm, picfn);
      }
    }
  }
}


void delete_oldpics(Pic *pics, char line[], FILE *tfp)
/* Check all the titles and filenames in the pics list against the
   titles and filename in the <li> line. If there is a match, do
   not output the line. */
{
  int found = 0;

  while ((pics != NULL) && (!found))
    if ((strstr(line, pics->fullfn) != NULL) ||
        (strstr(line, pics->title) != NULL))
      found = 1;
    else
      pics = pics->nextpic;
  if (!found)
    fputs(line, tfp);
}


void insert_newpics(Pic *pics, FILE *tfp)
/* Output the contents of the pics list as <li> lines. */
/* A line has the form:
     <li><a href="pics->fullfn#pic->lineno">pics->title</a> <p>
*/
{
  while (pics != NULL) {
    fprintf(tfp, "<li> <a href=\"%s#%d\">%s</a> <p>\n",
                       pics->fullfn, pics->lineno, pics->title);
    pics = pics->nextpic;
  }
}


void print_pics(Pic *pcs)
/* print utility for debugging pic d.s */
{
  Pic *p;
  p = pcs;
  while (p != NULL) {
    printf(" title is: %s\n", p->title);
    printf(" fullfn is: %s\n", p->fullfn);
    printf(" lineno is: %d\n\n", p->lineno);
    p = p->nextpic;
  }
  printf("\n");
}


/* Update NEWHTML file */

void mod_newfile(Entry *top)
/* Update the NEWHTML file with the current list of
   NEW pages. These are determined by having a '*' 
   against their names in the contents file.
*/
{
  FILE *fp, *tempfp;
  char line[MAXLINE], tmpnm[PATHLEN];
  int c;

  if (top != NULL) {
    if ((fp = fopen(NEWHTML,"r")) == NULL)
      printf("Cannot open %s\n", NEWHTML);
    else {
      strcpy(tmpnm, NEWHTML);
      strcat(tmpnm, "tmp");
      if ((tempfp = fopen(tmpnm,"w")) == NULL) {
        printf("\nCannot open temporary file for %s\n", NEWHTML);
        remove(tmpnm);
        fclose(fp);
      }
      else {
        printf("Modifying %s\n", NEWHTML);
        fgets(line, MAXLINE, fp);
        while (same_chars(line, 0, "<ul>") == 0) {
          fputs(line, tempfp);
          fgets(line, MAXLINE, fp);
        }
        fputs(line, tempfp);         /* skip over <ul> line */
        fgets(line, MAXLINE, fp);
        while (same_chars(line, 0, "</ul>") == 0) {
          fgets(line, MAXLINE, fp);
        }
        insert_newpgs(top, tempfp);
        fputs(line, tempfp);       /* write out the </ul> line */
        while ((c = getc(fp)) != EOF)
          putc(c, tempfp);

        fclose(fp); fclose(tempfp);
        remove(NEWHTML);
        rename(tmpnm, NEWHTML);
      }
    }
  }
}


void insert_newpgs(Entry *et, FILE *tfp)
/* Each new page is listed. Each line includes a URL and
   the title of the page (if one exists).
*/
{
  char tstring[MAXLINE];
  char ffn[PATHLEN];

  while (et != NULL) {
    if (et->new == 1) {
      sprintf(ffn, "%s%s.html", et->path->wd, et->name->wd);
      read_title(et, ffn, tstring);
      fprintf(tfp, "<li> <a href=\"%s\">%s</a><p>\n", ffn, tstring);
    }
    et = et->nextentry;
  }
}


/* redefine difftime() since SunOS doesn't have it! */
double mydifftime(time_t t1, time_t t0)
{
  return (double) (long) t1 - (long) t0;
}


/* Strings manipulation */

Strings strsbld(char *s)
/* Insert s into a list hanging from a cell of strs_store[].
   This is chosen by looking at the first letter of s.
*/
{
  return strs_insert( &strs_store[strs_hash(s)], s);
}

int strs_hash(char *s)
{
  if (isalpha(s[0]))
    return (tolower(s[0]) - 'a');
  else
    return 26;     /* catch-all spot */
}

Strings strs_insert(Strings *list, char *s)
/* Insertion is not sorted, but duplicates are avoided.
   If a duplicate is found, a pointer to that is returned,
   which avoids string space blow-out.
*/
{
  Strings curr, prev, new;
  int cmp = -1;  

  prev = NULL;
  curr = *list;

  while ((curr != NULL) &&
         ((cmp = strcmp(s, curr->wd)) != 0)) {
    prev = curr;
    curr = curr->next;
  }
  if ((prev == NULL) && (cmp != 0)) {
    new = mk_strs(s);
    new->next = *list;
    *list = new;
    return new;
  }
  if (curr == NULL) {
    new = mk_strs(s);
    prev->next = new;
    return new;
  }
  return curr;
}


Strings mk_strs(char *s)
{
  Strings temp;
  
  temp = (Strings)malloc(sizeof(Strnode));
  temp->wd = (char *)malloc(sizeof(char) * (strlen(s)+1));
  strcpy(temp->wd, s);
  temp->next = NULL;
  return temp;
}


void init_strs(void)
/* Fill strs_store[] cells with NULL, and add
   four useful strings immediately. */
{
  int x;
  for (x=0; x < STRSSIZE; x++)
    strs_store[x] = NULL;
  
  NIX = strsbld("");
  KH = strsbld(KEYSHTML);
  FH = strsbld(FIGSHTML);
  TH = strsbld(TBLSHTML);
  SH = strsbld(SRCHHTML);
  PH = strsbld(NEWHTML);
}


/* Debugging functions for printing out strs_store[] */

void print_strs(void)
{
  int x;

  printf("\nstrs_store contents\n");
  for (x=0; x < STRSSIZE; x++){
    printf("\nstrs_store[%d] contents:\n", x);
    print_str(strs_store[x]);
  }
}

void print_str(Strings list)
{
  while (list != NULL) {
    printf("%s, ", list->wd);
    list = list->next;
  }
  printf("END\n");
}

