
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
     decurl [file|directory]

   Date: May 1st, 1996

   Removes the header and footer information from a HTML file
   and also strips out any contents and replaces it with 
   %cnts/%fcnts. 
   It does the same job as write_middle() in curl.c 
   If supplied with a directory, it recursively applies itself to all
   the files in the tree. 

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 287 9172
   Fax: +61 3 348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

/* CHANGES

3/1/96. Modified find_body() to accept <body ... and </body>
(ad@cs.mu.oz.au)

*/


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ftw.h>

#define MAXLINE 200
#define PATHLEN	100

int delink(const char *fname, const struct stat *status, int type);
void strip(const char *fname);
int write_middle(const char *nm, FILE *tfp, FILE *hfp);
void find_body(const char *nm, FILE *fp, int *start, int *end);
int same_chars(char line[], int index, char pat[]);
void strip_pic(char line[], char nline[]);
int last_header(FILE *fp);
int html_cmp(const char *fname);

int main(int argc, char *argv[])
{
  struct stat stbuf;

  if (argc == 1)
    ftw(".", delink, 1);     /* using current directory */
  else if (argc == 2) {
      if (stat(argv[1], &stbuf) == -1) {
        printf("Cannot read stat structure for %s\n", argv[1]);
        return (-1);
      }
      if ((stbuf.st_mode & S_IFMT) == S_IFDIR)  /* argv[1] is dir */
        ftw(argv[1], delink, 1); 
      else {
        if (html_cmp(argv[1]) == 0)
          strip(argv[1]);        /* argv[1] is a HTML file */
        else
          printf("Ignoring %s since it does not end with .html\n", argv[1]);
      }
  }
  else
    printf("Wrong number of arguments. Usage: decurl [file|directory]\n");
  return 0;
}


int delink(const char *fname, const struct stat *status, int type)
/* If fname is a HTML file then apply strip() to it. */
/* delink is called by ftw() for every object in a directory.
   See p.74-76 of UNIX System Programming */
{
  if (type == FTW_F)                   /* fname is a file */
    if (html_cmp(fname) == 0)     /* fname is a HTML file */
       strip(fname);
  /*  else
        printf("Ignoring %s since it does not end with .html\n", fname); */
  return 0;
}


void strip(const char *fname)
/* strip off the extra bits from fname by writing to
   a temporary file (add "tmp" to the end). Then overwrite
   original file.  */
{
  char tempnm[PATHLEN];
  FILE *fp, *tfp;
  int last;
  
  if ((fp = fopen(fname,"r")) != NULL) {
    last = last_header(fp);
    if (last == -1) {
      printf("Ignoring %s, since no curl header found\n", fname);
      fclose(fp);
    }
    else {
      strcpy(tempnm, fname);
      strcat(tempnm, "tmp");
      if ((tfp = fopen(tempnm, "w")) != NULL) {
        printf("Decurling %s\n", fname);
        if (write_middle(fname, tfp, fp) == 1) {  /* see ts2html code */
          fclose(tfp); fclose(fp);
          remove(fname);
          rename(tempnm, fname);   /* move text in temp to html */
        }
        else {   /* cannot strip out curl info */
          fclose(tfp); fclose(fp);
          remove(tempnm);    /* get rid of temp file */
        }
      }
      else {
        printf("Cannot open a temporary file for %s\n", fname);
        fclose(fp);
      }
    }
  }
  else
    printf("Cannot open %s\n", fname);
}


int write_middle(const char *nm, FILE *tfp, FILE *hfp)
/* Extracts the textual 'middle' of a HTML file.
   This means stripping off the header and footer parts.
   Also any cnts/fcnts entries are replaced by %cnts/%fcnts labels
   so that they can be reprocessed later.
   The links attached to Figures and Tables are also removed.
*/
{
  char line[MAXLINE], nline[MAXLINE];
  int i = 0;
  int footflag = 0;
  int bodystart = -1;
  int bodyend = -1;

  find_body(nm, hfp, &bodystart, &bodyend);
  if (bodystart == -1) {
    printf("%s being ignored since it has no <body> tag\n", nm);
    return -1;
  }
  if (bodyend == -1) {
    printf("%s being ignored since it has no </body> tag\n", nm);
    return -1;
  }
  while (i < bodystart) {    /* skip to start of body */
    fgets(line, MAXLINE, hfp);
    i++;
    fputs(line, tfp);
  }

  while (strcmp(line, "<!-- end-of-header -->\n") != 0) {
    fgets(line, MAXLINE, hfp);
    i++;
  }

  while (i < bodyend) {
    fgets(line, MAXLINE, hfp);
    i++;
    if (strcmp(line, "<!-- start-of-cnts -->\n") == 0) {
      while (strcmp(line, "<!-- end-of-cnts -->\n") != 0)
        if (fgets(line, MAXLINE, hfp) == NULL) {
          printf("Missing <!-- end-of-cnts -->\n");
          return -1;
        }
        else i++;
      fprintf(tfp, "%ccnts\n", '%');          /* want %cnts printed */
    }
    else if (strcmp(line, "<!-- start-of-fcnts -->\n") == 0) {
      while (strcmp(line, "<!-- end-of-fcnts -->\n") != 0)
        if (fgets(line, MAXLINE, hfp) == NULL) {
          printf("Missing <!-- end-of-fcnts -->\n");
          return -1;
        }
        else i++;
      fprintf(tfp, "%cfcnts\n", '%');         /* want %fcnts printed */
    }
    else if (same_chars(line, 0, "<b>Figure</b>:") == 1) {
      strip_pic(line, nline);
      fputs(nline, tfp);
    }
    else if (same_chars(line, 0, "<b>Table</b>:") == 1) {
      strip_pic(line, nline);
      fputs(nline, tfp);
    }
    else if (strcmp(line, "<!-- start-of-footer -->\n") == 0) {
      footflag = 1;
      while (i < bodyend) {
        fgets(line, MAXLINE, hfp);
        i++;
      }
      fputs(line, tfp);
    }
    else
      fputs(line, tfp);
  }
  if (!footflag)      /* no footer encountered */
    printf("  Footer information missing\n");
  while (fgets(line, MAXLINE, hfp) != NULL)
    fputs(line, tfp);
  return 1;
}


void find_body(const char *nm, FILE *fp, int *start, int *end)
/* Set start to be the line number containing <body ...
   (don't look for '>' since there may be atttributes);
   end is assigned the line number containing </body>.
   They are -1 by default.
   The tags need to start the line.
*/
{
  char line[MAXLINE];
  int lineno = 0;

  while (fgets(line, MAXLINE, fp) != NULL) {
    lineno++;
    if ((same_chars(line, 0, "<body") == 1) ||
        (same_chars(line, 0, "<BODY") == 1)) {
      *start = lineno;
      break;
    }
  }
  if (*start != -1)
    while (fgets(line, MAXLINE, fp) != NULL) {
      lineno++;
      if ((same_chars(line, 0, "</body>") == 1) ||
          (same_chars(line, 0, "</BODY>") == 1)) {
      *end = lineno;
      break;
      }
    }
  rewind(fp);
}


int same_chars(char line[], int index, char pat[])
/* Is pat in line starting at index? */
{
  int i;

  for (i=0; pat[i] != '\0'; i++)
    if (line[index+i] != pat[i])
      return 0;
   return 1;
}


void strip_pic(char line[], char nline[])
/* Build a new line where the pic information after the
   first '.' has been deleted, and replaced by <p>. */
{
  int k = 0;

  while (line[k] != '.') {
    nline[k] = line[k];
    k++;
  }
  nline[k] = '\0';
  strcat(nline, ". <p>\n");
}


int last_header(FILE *fp)
/* Scan the entire file, and record the last line
   which has a "<!-- end-of-header -->". */
{
  char line[MAXLINE];
  int lineno, last;

  lineno = 0;
  last = -1;
  while (fgets(line, MAXLINE, fp) != NULL) {
    lineno++;
    if (strcmp(line, "<!-- end-of-header -->\n") == 0)
      last = lineno;
  }
  rewind(fp);
  return last;
}


int html_cmp(const char *fname)
/* Compare the file name end with "html" */
{
  int i=0;     

  if (fname[0] == '.') 
    i = 1;       /* to skip possible "." at beginning */
  while ((fname[i] != '\0') && (fname[i] != '.'))
    i++;
  if (fname[i] == '.')
    return strcmp(&fname[i+1], "html");
  else
    return 1;
}

