
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* sk.c 

   Date: May 1st, 1996

   Called as an CGI script by the search.html form.
   The program is passed the following:
        path url_path keyword1 keyword2 ... keywordMAXKEYS

   Recursively searches from the 'path' directory 
   and its subdirectories to see if any of the
   ".html" files contain the keywords.

   The matching files/WWW pages are ordered, and displayed.
   Only a maximum of LISTMAX matches are shown,
   but the total number of matches is also printed.
   The output includes the URL of the page (derived
   from the 'url_path' argument), and the <title> of the
   page (if there is one). The number of matches for each
   keyword in that file is also output.

   The file search uses fgrep with the -i option to ignore case
   and -c to return only a count of the lines that contain each
   keyword.

   Based on the directory listings example in
   K&R, section 8.6 (2nd ed.).
   HTML utilities written by Rob McCool.

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>        /* flags for read and write */
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>       /* directory features */


#define LISTMAX 30        /* maximum number of matches listed */
#define BIGMATCH 3        
            /* number of lines with a keyword that signals importance */
#define MAX_PATH 1024     /* max length of a full filename */
#define MAXKEYS 5         /* max number of keywords possible */
#define READLNS 30        /* max lines read for keyword/string/TITLE */
#define MAXLINE 1000      /* max length of input line */


typedef struct match {    /* data structure for a matching page */
  char *name; 
  int kscore[MAXKEYS];
  int mult, total;
  struct match *next;
}  Match;

typedef struct {          /* for the HTML name-value pairs */
    char *name;
    char *val;
} entry;


void check_env();
int fill_fields(char *fs[]);
void search_key(char *name, char *path, char *keys[], int kno, 
                                    Match **mlist, int *listno);
void dirwalk(char *dir, char *path, char *keys[], int kno, 
                                    Match **mlist, int *listno);
void search_file(char *name, char *path, char *keys[], int kno, 
                                    Match **mlist, int *listno);
int html_cmp(char *fname);
int worth_adding(Match *el, int kno);
void store_match(Match *el, Match **mlist);
void chop_last(Match *mlist);
void print_matches(Match *mlist, char *path, char *upath, int kno);
void read_title(char *name, char *path, char string[]);
int find_title(char *name, char line[], char string[]);
int start_title(char line[], int i);
int end_title(char line[], int i);
int same_chars(char line[], int index, char pat[]);

char *makeword(char *line, char stop);
char *fmakeword(FILE *f, char stop, int *len);
void unescape_url(char *url);
char x2c(char *what);
void plustospace(char *str);


int main()
{
  int fnum, i, kno;
  Match *mlist = NULL;
  int fmatches = 0;
  char *keys[MAXKEYS];
  char *fields[MAXKEYS+2];

  printf("Content-type: text/html\n\n");
  printf("<html><head>\n");
  printf("<title>Search Results</title>\n");
  printf("</head><body>\n");
  printf("<h1>Search Results</h1>\n");

  check_env();

  fnum = fill_fields(fields);

  for (i=2; i < fnum; i++)
    keys[i-2] = fields[i];     /* point to relevant field string */
  kno = fnum-2;

  printf("Search for the keyword(s):");
  for (i=0; i < kno; i++)
    printf(" \"%s\"", keys[i]);
  printf("<p>\n");
  printf("The bracketed numbers are the number of matching lines on each page for each keyword.<p><br>\n");
  search_key("", fields[0], keys, kno, &mlist, &fmatches);   
                                 /* fields[0] is the dir path */

  print_matches(mlist, fields[0], fields[1], kno);  
                                 /* fields[1] is the URL path */
  printf("<br>Total number of matching pages: %d<p>\n", fmatches);
  printf("</body></html>\n");
  return 0;
}



void check_env(void)
/* Check the REQUEST_METHOD and CONTENT_TYPE enviroment variables */
{
  if(strcmp(getenv("REQUEST_METHOD"),"POST")) {
    printf("<b>Error</b>: This script should be referenced with a METHOD of POST.<p>\n");
    printf("If you don't understand this, read ");
    printf("<A HREF=\"http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/fill-out-forms/overview.html\">forms overview</A>.<p>\n");
    printf("</body></html>\n");
    exit(-1);
  }
  if(strcmp(getenv("CONTENT_TYPE"),"application/x-www-form-urlencoded")) {
    printf("<b>Error</b>: This script can only be used to decode form results.<p>\n");
    printf("</body></html>\n");
    exit(-1);
  }
}


int fill_fields(char *fs[])
/* Read the HTML form data into the entries[] array, and then
   copy the non-empty fields into the fields[] array.
   fields[0] will be the path to the directory holding 
   the search.html file. fields[1] will be the URL path to
   search.html. The other fields (MAXKEYS of them) will be
   the keywords entered by the user.
*/
{
  int cl, i;
  int num = 0, fnum = 0;
  entry entries[MAXKEYS+2];

  cl = atoi(getenv("CONTENT_LENGTH"));

  for(i=0; cl && (!feof(stdin)); i++) {
    entries[i].val = fmakeword(stdin,'&',&cl);
    plustospace(entries[i].val);
    unescape_url(entries[i].val);
    entries[i].name = makeword(entries[i].val,'=');
    num++;
  }

  for (i=0; i < num; i++) {
    if (entries[i].val[0] != '\0') {
      fs[fnum] = (char *)malloc(sizeof(char)*(strlen(entries[i].val)+1));
      strcpy(fs[fnum], entries[i].val);
      fnum++;
    }
    free(entries[i].name);
    free(entries[i].val);
  }

  if (fnum < 3) {
    printf("<b>Error</b>: You must supply at least one keyword<p>\n");
    printf("</body></html>\n");
    exit (-1);
  }

  return fnum;
}



void search_key(char *name, char *path, char *keys[], int kno, 
                                 Match **mlist, int *fmatches)
/* Recursively search for matching files, storing the matches
   in mlist.
*/
{
  struct stat stbuf;
  char ffn[MAX_PATH];

  sprintf(ffn, "%s/%s", path, name);
  if (stat(ffn, &stbuf) == -1) {
    printf("<b>Error</b>: Cannot read stat structure for %s<p>\n", ffn);
    return;
  }
  if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
    dirwalk(name, path, keys, kno, mlist, fmatches);
  else if (html_cmp(ffn) == 0)
    search_file(name, path, keys, kno, mlist, fmatches);
}


void dirwalk(char *dir, char *path, char *keys[], int kno, 
                                Match **mlist, int *fmatches)
/* Process each file/subdirectory in a directory */
{
  char fdir[MAX_PATH];
  char name[MAX_PATH];
  struct dirent *dp;
  DIR *dfd;

  sprintf(fdir, "%s/%s", path, dir);
  if ((dfd = opendir(fdir)) == NULL) {
    printf("<b>Error</b>: Cannot open the directory %s<p>\n", fdir);
    return;
  }
  while ((dp = readdir(dfd)) != NULL) {
    if ((strcmp(dp->d_name, ".") == 0)
        || (strcmp(dp->d_name, "..") == 0))
      continue;   /* skip self and parent */
    if ((strlen(fdir) + strlen(dp->d_name) + 2) >= MAX_PATH)
      printf("<b>Error</b>: Name %s/%s too long<p>\n", fdir, dp->d_name);
    else {
      sprintf(name, "%s/%s", dir, dp->d_name);
      search_key(name, path, keys, kno, mlist, fmatches);
    }
  }
  closedir(dfd);
}


void search_file(char *name, char *path, char *keys[], int kno, 
                               Match **mlist, int *fmatches)
/* For a given file, build a match structure and record the
   number of matching lines for each keyword. If the match is
   worth adding then save it in mlist.
   Matching is done by calling fgrep.
*/
{
  char ffn[MAX_PATH];
  char cmd[MAX_PATH];
  int tot_hits = 0, mult_hits = 1, hits, i;
  FILE *strm;
  Match *temp;

  temp = (Match *) malloc(sizeof(Match));
  temp->name = (char *) malloc(sizeof(char)*(strlen(name)+1));
  strcpy(temp->name, name);
  temp->next = NULL;

  sprintf(ffn, "%s/%s", path, name);
  for (i=0; i < kno; i++){
    if ((strlen(keys[i]) + strlen(ffn) + 14) >= MAX_PATH) {
      printf("<b>Error</b>: \"fgrep -c -i %s %s\" command too long<p>\n", 
                                                      keys[i], ffn);
      free(temp->name);
      free(temp);
      return;
    }
    sprintf(cmd, "fgrep -c -i %s %s", keys[i], ffn);
  
    if ((strm = popen(cmd, "r")) == NULL) {
      printf("<b>Error</b>: \"fgrep -c -i %s %s\" command start error<p>\n", 
                                                   keys[i], ffn);
      free(temp->name);
      free(temp);
      return;
    }
    fscanf(strm, "%d", &hits);
    pclose(strm);
    temp->kscore[i] = hits;
    tot_hits = tot_hits + hits;
    mult_hits = mult_hits * hits;
  }
  temp->mult = mult_hits;
  temp->total = tot_hits;
  if (worth_adding(temp, kno) == 1) {
    (*fmatches)++;
    store_match(temp, mlist);
    if (*fmatches > LISTMAX)
      chop_last(*mlist);
  }
  else {
    free(temp->name);
    free(temp);
  }
}


int html_cmp(char *fname)
/* Compare the file name end with "html" */
{
  int i=0;

  if (fname[0] == '.')
    i = 1;       /* to skip possible "." at beginning */
  while ((fname[i] != '\0') && (fname[i] != '.'))
    i++;
  if (fname[i] == '.')
    return strcmp(&fname[i+1], "html");
  else
    return 1;
}


int worth_adding(Match *el, int kno)
/* A few criteria are applied:
    1. If there are no matches then it isn't important.
    2. If one keyword has >= BIGMATCH matches then its
       worth adding (even if some of the other keywords
       haven't a single match.
    3. If the number of zero matches is over half of the
       total number of keywords then it isn't important.
*/
{
  int i, numzs = 0;

  if (el->total == 0)
    return 0;

  for (i=0; i < kno; i++)
    if (el->kscore[i] >= BIGMATCH)
      return 1;

  for (i=0; i < kno; i++)
    if (el->kscore[i] == 0)
      numzs++;
  if ((numzs*2) > kno)
    return 0;

  return 1;
}


void store_match(Match *el, Match **mlist)
/* The insertion strategy uses two increasing sorts.
   The first sort is obtained by multiplying the line counts 
   for each keyword together.
   The second sort is obtained by summing the line counts 
   for each keyword. 
*/
{
  Match *prev, *curr;

  prev = NULL;
  if (*mlist == NULL) {
    *mlist = el;
    return;
  }

  curr = *mlist;
  while ((curr != NULL) && (curr->mult > el->mult)) {
    prev = curr;
    curr = curr->next;
  }

  if (curr == NULL)
    prev->next = el;
  else if (el->mult > curr->mult) {
    el->next = curr;
    if (prev != NULL)
      prev->next = el;
    else
      *mlist = el;
  }
  else {
    while ((curr != NULL) && (curr->total > el->total)) {
      prev = curr;
      curr = curr->next;
    }
    if (curr == NULL)
      prev->next = el;
    else {
      el->next = curr;
      if (prev != NULL)
        prev->next = el;
      else
       *mlist = el;
    }
  }
}


void chop_last(Match *mlist)
/* Free the last element of mlist. This will start
   happening after LISTMAX elements have been inserted.
*/
{
  while (mlist->next->next != NULL)
    mlist = mlist->next;
  free(mlist->next->name);
  free(mlist->next);
  mlist->next = NULL;
}


void print_matches(Match *mlist, char *path, char *upath, int kno)
/* Display the list of matches as a bulleted <ul> list.
   Each match will include the URL of the page, its title,
   and the number of matching lines for each keyword.
*/
{
  int i, x=0;
  char title[MAXLINE];

  printf("<ul>\n");
  while ((mlist != NULL) && (x != LISTMAX)) {
    printf("<li> ");
    read_title(mlist->name, path, title);
    printf("<a href=\"%s%s\">%s</a>.", upath, mlist->name, title);
    printf(" (");
    for (i=0; i < kno; i++)
      printf(" %d", mlist->kscore[i]);
    printf(" )\n");
    mlist = mlist->next;
    x++;
  }
  printf("</ul> <p>\n");
}



void read_title(char *name, char *path, char ttle[])
/* The <TITLE> line is looked for in the first READLNS lines of
   a file. The format is:
   <TITLE>string-bits</TITLE>
   If there isn't a TITLE line then the the full filename (ffn)
   is used as the title.
*/
{
  FILE *fp;
  char ffn[MAX_PATH];
  char line[MAXLINE];
  int i, titlefound = 0;

  sprintf(ffn, "%s%s", path, name);
  if ((fp = fopen(ffn,"r")) == NULL) {
    printf("Cannot open %s to read TITLE line\n", ffn); 
    strcpy(ttle, ffn);
  }
  else {
    for (i=0; ((i < READLNS) && (fgets(line, MAXLINE, fp) != NULL)
                             && (!titlefound)); i++)
      if (find_title(ffn, line, ttle))
        titlefound = 1;
    fclose(fp);
    if (!titlefound)
      strcpy(ttle, ffn);
  }
}


int find_title(char *ffn, char line[], char ttle[])
/* The <TITLE> line in a file has the form:
      <TITLE>string-bit</TITLE>
   or <title>string-bit</title>
*/
{
  int i, k;

  if (start_title(line, 0) == 1) {
    i = 7;
    k = 0;
    while ((end_title(line, i) == 0) && (i < MAXLINE))
      ttle[k++] = line[i++];
    ttle[k] = '\0';
    if (i == MAXLINE) {
      printf("In %s, Could not find </TITLE> in line:   %s", ffn, line);
      return 0;
    }
    else
      return 1;
  }
  return 0;
}


int start_title(char line[], int i)
/* check for <TITLE> or <title> */
{
  return (same_chars(line, i, "<TITLE>") ||
          same_chars(line, i, "<title>"));
}


int end_title(char line[], int i)
/* check for </TITLE> or </title> */
{
  return (same_chars(line, i, "</TITLE>") ||
          same_chars(line, i, "</title>"));
}


int same_chars(char line[], int index, char pat[])
/* Is pat in line starting at index? */
{
  int i;

  for (i=0; pat[i] != '\0'; i++)
    if (line[index+i] != pat[i])
      return 0;
   return 1;
}


/* HTML utilities */

char *makeword(char *line, char stop) {
  int x = 0,y;
  char *word = (char *) malloc(sizeof(char) * (strlen(line) + 1));

  for(x=0;((line[x]) && (line[x] != stop));x++)
    word[x] = line[x];

  word[x] = '\0';
  if(line[x]) ++x;
  y=0;

  while((line[y++] = line[x++]) != 0);
  return word;
}

char *fmakeword(FILE *f, char stop, int *cl) {
  int wsize;
  char *word;
  int ll;

  wsize = 102400;
  ll=0;
  word = (char *) malloc(sizeof(char) * (wsize + 1));

  while(1) {
    word[ll] = (char)fgetc(f);
    if(ll==wsize) {
        word[ll+1] = '\0';
        wsize+=102400;
        word = (char *)realloc(word,sizeof(char)*(wsize+1));
    }
    --(*cl);
    if((word[ll] == stop) || (feof(f)) || (!(*cl))) {
        if(word[ll] != stop) ll++;
        word[ll] = '\0';
        return word;
    }
    ++ll;
  }
}

void unescape_url(char *url) {
  register int x,y;

  for(x=0,y=0;url[y];++x,++y) {
    if((url[x] = url[y]) == '%') {
        url[x] = x2c(&url[y+1]);
        y+=2;
    }
  }
  url[x] = '\0';
}

char x2c(char *what) {
  register char digit;

  digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
  digit *= 16;
  digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
  return(digit);
}

void plustospace(char *str) {
  register int x;

  for(x=0;str[x];x++)
    if(str[x] == '+') str[x] = ' ';
}


