
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
     delink links-file

   Date: May 1st, 1996

   This program modifies a series of HTML files so that
   the links previously added by "linkup" are removed.

   The linkup program should be called from the 'top' 
   directory of the WWW presentation containing the HTML
   files. 

   The links-file must contain the following:
      - url address of this 'top' directory (ending in a '/')
      - a line used to label the collection of links
        added to each file
      - a series of relative paths from the home directory to the files,
        each on its own line

   Blank lines are allowed.
   Not all of this is necessary, but it allows the same link file
   format to be used for both "linkup" and "delink".

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINE 1000      /* max length of input line */
#define MAX_NAMES 10      /* max number of files to be inter-linked */
#define TMPFILE "temp$$"  /* temporary file name */


int link_info(FILE *fp, char **top_url, char **link_label, char *fnames[]);
int get_line(char **pstr, FILE *fp);
int same_chars(char line[], int index, char pat[]);
void delink_files(char *fnames[], int num);
int rem_links(FILE *fp, FILE *tfp, char *fnm);


int main(int argc, char *argv[])
{
  char *top_url;
  char *link_label;
  char *fnames[MAX_NAMES];
  int num_names;
  FILE *fp;

  if (argc != 2) {
    printf("Usage: delink links-file\n");
    exit(1);
  }
  if ((fp = fopen(argv[1], "r")) == NULL) {
    printf("Cannot open links-file: %s\n", argv[1]);
    exit(1);
  }

  printf("Reading links-file: %s\n", argv[1]);
  num_names = link_info(fp, &top_url, &link_label, fnames);
  fclose(fp);

  if (num_names > 0)
      delink_files(fnames, num_names);

  return 0;
}


int link_info(FILE *fp, char **top_url, char **link_label, char *fnames[])
/* Extract the info from the links file */
{
  int i, num = 0;

  if (get_line(top_url, fp)) {
    if (get_line(link_label, fp)) {
      while ((get_line(&fnames[num], fp)) && (num < MAX_NAMES))
        num++;
      if (num == MAX_NAMES)
        printf("%d file names read; others ignored\n", MAX_NAMES);

      printf("Top URL: %s\n", *top_url);
      printf("Link Label: %s\n", *link_label);
      printf("File names:\n  ");
      for (i=0; i < num; i++)
        printf("%s  ", fnames[i]);
      printf("\n\n");
    }
  }
  return num;
}


int get_line(char **pstr, FILE *fp)
/* Get the first non-blank line and make pstr point to it.
   If no suitable line is found return 0
*/
{
  int found_line = 0, len;
  char line[MAXLINE];

  while (!found_line) {
    if (fgets(line, MAXLINE, fp) == NULL)
      return 0;
    if (line[0] != '\n')
      found_line = 1;
  }

  len = strlen(line);
  *pstr = (char *) malloc(sizeof(char)*len);
  strncpy(*pstr, line, len);
  (*pstr)[len-1] = '\0';    /* overwrite '\n' */

  return 1;
}



int same_chars(char line[], int index, char pat[])
/* Is pat in line starting at index? */
{
  int i;

  for (i=0; pat[i] != '\0'; i++)
    if (line[index+i] != pat[i])
      return 0;
   return 1;
}


void delink_files(char *fnames[], int num)
/* For each file, build a temporary file with the links
   removed. Then make the temp file the new version of the original.
*/
{
  int i;
  FILE *fp, *tfp;

  for (i=0; i < num; i++) {
    if ((fp = fopen(fnames[i],"r")) == NULL) {
      printf("Cannot open %s for reading\n", fnames[i]);
      exit(1);
    }     
    if ((tfp = fopen(TMPFILE, "w")) == NULL) {
      printf("Cannot open temporary file %s for writing\n", TMPFILE);
      exit(1);
    }
    printf("Removing links from %s\n", fnames[i]);
    
    if (rem_links(fp, tfp, fnames[i]) == 1) {
      fclose(fp);
      fclose(tfp);
      remove(fnames[i]);
      rename(TMPFILE, fnames[i]);
    }
    else {
      printf("Delinking of %s aborted\n", fnames[i]);
      fclose(fp);
      fclose(tfp);
      remove(TMPFILE);
    }
  }
}

 
int rem_links(FILE *fp, FILE *tfp, char *fnm)
/* Find the <!-- start of links --> tag line, and remove all
   the lines down to the <!-- end of links --> line.
*/
{
  char line[MAXLINE];
  int start_links = 0, end_links = 0;

  while (fgets(line, MAXLINE, fp) != NULL) {
    if ((start_links = same_chars(line, 0, "<!-- start of links -->")) == 1)
      break;
    fputs(line, tfp);
  }

  if (!start_links) {
    printf("No <!-- start of links --> tag in %s\n", fnm);
    return 0;
  }

  while (fgets(line, MAXLINE, fp) != NULL)
    if ((end_links = same_chars(line, 0, "<!-- end of links -->")) == 1)
      break;

  if (!end_links) {
    printf("No <!-- end of links --> tag in %s\n", fnm);
    return 0;
  }

  while (fgets(line, MAXLINE, fp) != NULL)
    fputs(line, tfp);

  return 1;
}

