
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
     linkup links-file

   Date: May 1st, 1996

   This program modifies a series of HTML files so that
   they all have relative URL links to each other.

   The linkup program should be called from the 'top' 
   directory of the WWW presentation containing the HTML
   files. 

   The links-file must contain the following:
      - url address of this 'top' directory (ending in a '/')
      - a line used to label the collection of links
        added to each file
      - a series of relative paths from the home directory to the files,
        each on its own line

   Blank lines are allowed.
   The links are added just before the curl footer tag or the
   </body> (</BODY>) tag, and are preceded by the comment line:
     <!-- start of links -->
   and followed by:
     <!-- end of links -->

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define READLNS 30        /* max lines read for keyword/string/TITLE */
#define MAXLINE 1000      /* max length of input line */
#define MAX_NAMES 10      /* max number of files to be inter-linked */
#define TMPFILE "temp$$"  /* temporary file name */


int link_info(FILE *fp, char **top_url, char **link_label, char *fnames[]);
int get_line(char **pstr, FILE *fp);
int read_title(char *fnm, char **ttl);
int find_title(char *fnm, char line[], char **ttl);
int start_title(char line[], int i);
int end_title(char line[], int i);
int same_chars(char line[], int index, char pat[]);
void link_files(char *fnames[], char *titles[], int num,
                           char *top_url, char *link_label);
void add_links(FILE *fp, FILE *tfp, char *link_label, char *top_url,
                 char *titles[], char *fnames[], int i, int num);
void place_list(FILE *tfp, char *top_url, char *fnames[],
                                    char *titles[], int i, int num);


int main(int argc, char *argv[])
{
  char *top_url;
  char *link_label;
  char *fnames[MAX_NAMES];
  char *titles[MAX_NAMES];
  int i, num_names, readable;
  FILE *fp;

  if (argc != 2) {
    printf("Usage: linkup links-file\n");
    exit(1);
  }
  if ((fp = fopen(argv[1], "r")) == NULL) {
    printf("Cannot open links-file: %s\n", argv[1]);
    exit(1);
  }

  printf("Reading links-file: %s\n", argv[1]);
  num_names = link_info(fp, &top_url, &link_label, fnames);
  fclose(fp);

  if (num_names > 0){
    readable = 1;
    for (i=0; i < num_names; i++)
      readable = readable && read_title(fnames[i], &titles[i]);
    if (readable)
      link_files(fnames, titles, num_names, top_url, link_label);
  }

  return 0;
}


int link_info(FILE *fp, char **top_url, char **link_label, char *fnames[])
/* Extract the info from the links file */
{
  int i, num = 0;

  if (get_line(top_url, fp)) {
    if (get_line(link_label, fp)) {
      while ((get_line(&fnames[num], fp)) && (num < MAX_NAMES))
        num++;
      if (num == MAX_NAMES)
        printf("%d file names read; others ignored\n", MAX_NAMES);

      printf("Top URL: %s\n", *top_url);
      printf("Link Label: %s\n", *link_label);
      printf("File names:\n  ");
      for (i=0; i < num; i++)
        printf("%s  ", fnames[i]);
      printf("\n\n");
    }
  }
  return num;
}   


int get_line(char **pstr, FILE *fp)
/* Get the first non-blank line and make pstr point to it.
   If no suitable line is found return 0
*/
{
  int found_line = 0, len;
  char line[MAXLINE];

  while (!found_line) {
    if (fgets(line, MAXLINE, fp) == NULL)
      return 0;
    if (line[0] != '\n')
      found_line = 1;
  }

  len = strlen(line);
  *pstr = (char *) malloc(sizeof(char)*len);
  strncpy(*pstr, line, len);
  (*pstr)[len-1] = '\0';    /* overwrite '\n' */

  return 1;
}



int read_title(char *fnm, char **ttl)
/* The <TITLE> line is looked for in the first READLNS lines of
   the file. The format is:
   <TITLE>string-bits</TITLE>
   If there isn't a TITLE line then the the filename
   is used as the title.
*/
{
  FILE *fp;
  char line[MAXLINE];
  int i, titlefound, readable = 1;

  if ((fp = fopen(fnm,"r")) == NULL) {
    printf("Cannot open %s to read TITLE line\n", fnm); 
    *ttl = (char *)malloc(sizeof(char)*(strlen(fnm)+1));
    strcpy(*ttl, fnm);
    readable = 0;
  }
  else {
    titlefound = 0;
    printf("Reading from %s\n", fnm);
    for (i=0; ((i < READLNS) && (fgets(line, MAXLINE, fp) != NULL)
                             && (!titlefound)); i++)
      titlefound = find_title(fnm, line, ttl);
    fclose(fp);
    if (!titlefound) {
      *ttl = (char *)malloc(sizeof(char)*(strlen(fnm)+1));
      strcpy(*ttl, fnm);
    }
  }
  if (readable)
    printf("Title: %s\n", *ttl);

  return readable;
}


int find_title(char *fnm, char line[], char **ttl)
/* The <TITLE> line in a file has the form:
      <TITLE>string-bit</TITLE>
   or <title>string-bit</title>
*/
{
  char title[MAXLINE];
  int i, k;

  if (start_title(line, 0) == 1) {
    i = 7;
    k = 0;
    while ((end_title(line, i) == 0) && (i < MAXLINE))
      title[k++] = line[i++];
    title[k] = '\0';
    if (i == MAXLINE) {
      printf("In %s, Could not find </TITLE> in line:\n   %s", fnm, line);
      return 0;
    }
    else {
      *ttl = (char *)malloc(sizeof(char)*(strlen(title)+1));
      strcpy(*ttl, title);
      return 1;
    }
  }
  return 0;
}


int start_title(char line[], int i)
/* check for <TITLE> or <title> */
{
  return (same_chars(line, i, "<TITLE>") ||
          same_chars(line, i, "<title>"));
}


int end_title(char line[], int i)
/* check for </TITLE> or </title> */
{
  return (same_chars(line, i, "</TITLE>") ||
          same_chars(line, i, "</title>"));
}


int same_chars(char line[], int index, char pat[])
/* Is pat in line starting at index? */
{
  int i;

  for (i=0; pat[i] != '\0'; i++)
    if (line[index+i] != pat[i])
      return 0;
   return 1;
}



void link_files(char *fnames[], char *titles[], int num, 
                           char *top_url, char *link_label)
/* For each file, build a temporary file with the links
   included. Then make the temp file the new version of the original.
*/
{
  int i;
  FILE *fp, *tfp;

  for (i=0; i < num; i++) {
    if ((fp = fopen(fnames[i],"r")) == NULL) {
      printf("Cannot open %s for reading\n", fnames[i]);
      exit(1);
    }     
    if ((tfp = fopen(TMPFILE, "w")) == NULL) {
      printf("Cannot open temporary file %s for writing\n", TMPFILE);
      exit(1);
    }
    printf("Adding links to %s\n", fnames[i]);
    add_links(fp, tfp, link_label, top_url, titles, fnames, i, num);
    fclose(fp);
    fclose(tfp);
    remove(fnames[i]);
    rename(TMPFILE, fnames[i]);
  }
}

 
void add_links(FILE *fp, FILE *tfp, char *link_label, char *top_url,
                 char *titles[], char *fnames[], int i, int num)
/* Find the curl footer tag or the /body tag, and start to add
   the list of links. These are embedded inside comments
   to aid their possible future removal.
*/
{
  char line[MAXLINE];
  int endbody = 0;

  while (fgets(line, MAXLINE, fp) != NULL) {
    if ((endbody = (same_chars(line, 0, "</body>") ||
                    same_chars(line, 0, "</BODY>") ||
                    same_chars(line, 0, "<!-- start-of-footer -->"))) == 1)
      break;
    fputs(line, tfp);
  }

  if (!endbody)
    printf("Could not find curl footer or /body tag in %s\n", fnames[i]);
  else {
    fprintf(tfp, "<!-- start of links -->\n");
    fprintf(tfp, "<P>%s:</P>\n", link_label);
    place_list(tfp, top_url, fnames, titles, i, num);
    fprintf(tfp, "<!-- end of links -->\n");
    fputs(line, tfp);              /* end body line */
    while (fgets(line, MAXLINE, fp) != NULL)
      fputs(line, tfp);
  }
}


void place_list(FILE *tfp, char *top_url, char *fnames[],
                                    char *titles[], int i, int num)
/* The links are added in the form of an unnumbered list,
   with one line for each of the other files.
   The URL is a combination of the top URL and the filename,
   and the title in the file is used as the URL label.
*/
{
  int j;

  fprintf(tfp, "<ul>\n");
  for (j=0; j < num; j++) 
    if (j != i)      /* this file */
      fprintf(tfp, "<li><a href=\"%s%s\">%s</a>\n", top_url, fnames[j],
                                               titles[j]);
  fprintf(tfp, "</ul><p>\n");
}


