
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
     occ_filt file min max

   e.g.    occ_filt occurs-list 8 19

   Date: May 1st, 1996

   The file is assumed to be in the format generated by the occurs.c
   program. This is filtered so that only words which have occurred
   between min and max times are output.

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAXLINE 200     /* max length of a line */
#define WSIZE 20        /* max length of a word */
#define COL_WIDTH 4     /* the number of columns printed */

char *process_args(int argc, char *argv[], int *min, int *max);
void filter_occurs(FILE *fp, int min, int max, char *fnm);
void filter_wo(char *line, int *i, char *alpha, int min, int max, int *cols);


int main(int argc, char *argv[])
{
  char *fnm;
  FILE *fp;
  int min, max;

  fnm = process_args(argc, argv, &min, &max);

  if ((fp = fopen(fnm, "r")) == NULL) {
    printf("Cannot open occurrences file: %s\n", fnm);
    exit(1);
  }
  else
    printf("Reading occurrences file: %s\n", fnm);
  filter_occurs(fp, min, max, fnm);
  fclose(fp);
  return 1;
}


char *process_args(int argc, char *argv[], int *min, int *max)
/* Extract the filename and the minimum and maximum values from
   the command line. */
{
  int temp;

  if (argc != 4) {
    printf("Usage: occ_filt file min max\n");
    exit(1);
  }
  *min = atoi(argv[2]);
  *max = atoi(argv[3]);
  if (*min < 0) *min = 0;
  if (*max < 0) *max = 0;
  if (*max < *min) {
    temp = *min;
    *min = *max;
    *max = temp;
  }
  printf("Minimum: %d; Maximum: %d\n", *min, *max);

  return argv[1];
}


void filter_occurs(FILE *fp, int min, int max, char *fnm)
/* Ignore all the lines until a line beginning with "Word"
   is reached (i.e. the "Word Occurrences" heading).
   Now read in each line and extract the:
      word: occurs 
   pairs on the line.
*/
{
  char line[MAXLINE], alpha;
  int found_occs = 0, i, cols;

  while ((fgets(line, MAXLINE, fp) != NULL) && (found_occs == 0))
    if (strcmp(line, "Word") > 0)
      found_occs = 1;

  if (found_occs == 0) {
    printf("%s does not contain any word occurrences\n", fnm);
  }
  else {
    cols = 0;
    alpha = '$';   /* some dummy value which isn't a letter */
    printf("\nWord Occurrences");
    while (fgets(line, MAXLINE, fp) != NULL) {
      i = 0;
      while (line[i] != '\n')
        if (isalpha(line[i]))
          filter_wo(line, &i, &alpha, min, max, &cols);
        else
          i++;
    }
    printf("\n\n");
  }
}


void filter_wo(char *line, int *i, char *alpha, int min, int max, int *cols)
/* Extract a single
      word: occurs
   pair from the line.
   Once the pair has been extracted, print it out only if occurs is
   within the min/max range.
   Print in columns.
   If the word is the first one of a new letter of the alphabet
   then print the letter.
*/
{
  char word[WSIZE], digit[WSIZE];
  int w, occurs;

  w = 0;
  while (isalpha(line[*i])) {
    word[w] = line[*i];
    w++; (*i)++;
  }
  word[w] = '\0';

  if (line[*i] == ':') {
    *i = *i + 2;   /* skip ':' and ' ' */
    w = 0;
    while (isdigit(line[*i])) {
      digit[w] = line[*i];
      w++; (*i)++;
    }
    digit[w] = '\0';
    occurs = atoi(digit);
  
    if (word[0] != *alpha) {
      *alpha = word[0];
      printf("\n\n%c\n", *alpha);
      printf("===\n");
      *cols = 0;
    }
    if ((occurs >= min) && (occurs <= max)) {
      printf("%10s: %d  ", word, occurs);
      if (*cols == (COL_WIDTH-1)) {
        putchar('\n');
        *cols = 0;
      }
      else
        (*cols)++;
    }
  }
}

