
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
      occurs [-s] [-m NUM] html_file | directory

   e.g.    occurs -m 3 "."
           occurs -s humour

   Date: May 1st, 1996

   Search through an HTML file or a directory. If the
   search is on a directory it will recursively search 
   all the subdirectories unless the -s option is set.

   The search collects words and records their number
   of occurrences. Words less than MINLEN in length are
   ignored, as are words loaded from an IGFNM file.

   The recorded words and their occurrences are printed at
   the end in alphabetical order. No distinction is made
   between upper and lower case.
  
   If a minimum NUM has been specified then only occurrences
   >= NUM are printed.

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>        /* flags for read and write */
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>       /* directory features */


#define NUM_WORDS 26     /* 26 for a-z */
#define WSIZE 20         /* the maximum size of a word */
#define MAX_PATH 1024     /* max length of a full filename */
#define MINLEN 3         /* minimum length of a word */
#define IGFNM "ignores.txt"   /* file of ignorable words */
#define WSENT "$"        /* sentinel word */
#define CSENT -1         /* sentinel count */
#define COL_WIDTH 4      /* the number of columns printed */

struct word_occ {
  char *word;
  int count;
  struct word_occ *next;
};

typedef struct word_occ Words;

char *process_args(int argc, char *argv[], int *sl, int *isf, int *mnum);
int html_cmp(char *fname);
void init_words(Words *ws[]);
Words *new_word(char *wd, int c);
void build_igs(char *fnm, Words *igs[]);
void make_lower(char wd[]);
void store_word(char *wd, Words **ws);
void rec_wds(char *fnm, Words *wds[], Words *igs[]);
void dirwalk(char *dir, int sing_lvl, Words *wds[], Words *igs[]);
void search_files(char *name, int sing_lvl, Words *wds[], Words *igs[]);
int ignore_me(char *wd, Words *ws);
void print_words(int mnum, Words *ws[]);
void print_occurrences(int mnum, Words *ws);


int main(int argc, char *argv[])
{
  int sing_lvl, is_hfile, mnum;
  char *srch_nm;
  Words *wds[NUM_WORDS];   /* for recorded words */
  Words *igs[NUM_WORDS];   /* for words to be ignored */

  srch_nm = process_args(argc, argv, &sing_lvl, &is_hfile, &mnum);
  init_words(wds);
  init_words(igs);
  build_igs(IGFNM, igs);

  if (is_hfile)
    rec_wds(srch_nm, wds, igs);
  else
    dirwalk(srch_nm, sing_lvl, wds, igs);

  print_words(mnum, wds);
  return 1;
}


char *process_args(int argc, char *argv[], int *sl, int *isf, int *mnum)
/* From the command line, extract the single level flag (sl), 
   the name of the file or directory (setting the flag isf in the process), 
   and the minimum occurrence number (mnum).
*/
{
  char *name;

  *sl =0;
  *mnum = 0;

  if ((argc > 5) || (argc == 1)) {
    printf("Usage: occurs [-s] [-m NUM] html_file | directory\n");
    exit(1);
  }

  if (strcmp(argv[1], "-s") == 0) {
    *sl = 1;
    if (strcmp(argv[2], "-m") == 0) {
      *mnum = atoi(argv[3]);
      name = argv[4];
    }
    else
      name = argv[2];
  }
  else if (strcmp(argv[1], "-m") == 0) {
    *mnum = atoi(argv[2]);
    name = argv[3];
  }
  else
    name = argv[1];

  if (*mnum > 0)
    printf("%d or more occurrences printed\n", *mnum);

  if (html_cmp(name) == 0) {
    *isf = 1;
    printf("Processing the HTML file \"%s\"\n", name);
  }
  else {
    *isf = 0;
    printf("Processing the directory \"%s\"", name);
    if (*sl == 1)
      printf(" ONLY\n\n");
    else
      printf(" AND all of its subdirectories\n\n");
  }
  return name;
}


int html_cmp(char *fname)
/* Compare the file name end with "html" */
{
  int i=0;

  if (fname[0] == '.')
    i = 1;       /* to skip possible "." at beginning */
  while ((fname[i] != '\0') && (fname[i] != '.'))
    i++;
  if (fname[i] == '.')
    return strcmp(&fname[i+1], "html");
  else
    return 1;
}



void init_words(Words *ws[])
/* Make each array element point to a sentinel struct */
{
  int i;
  for(i=0; i < NUM_WORDS; i++)
    ws[i] = new_word(WSENT, CSENT);  /* sentinel values */
}


Words *new_word(char *wd, int c)
/* Create a new word with a given word string (wd) and count (c) */
{
  Words *tmp;

  tmp = (Words *) malloc(sizeof(Words));
  tmp->word = (char *) malloc( sizeof(char) *
                               (strlen(wd)+1));
  strcpy(tmp->word, wd);
  tmp->count = c;
  tmp->next = NULL;
  return tmp;
}


void build_igs(char *fnm, Words *igs[])
/* Build the ignorable words array by reading words from a file */
{
  char wd[WSIZE];
  int posn;
  FILE *fp;

  if ((fp = fopen(fnm, "r")) == NULL)
    printf("Cannot read ignorable words file: %s\n", fnm);
  else {
    printf("Reading ignorable words file: %s\n", fnm);
    while (fscanf(fp, "%s", wd) != EOF) {\
      make_lower(wd);
      posn = wd[0] - 'a';
      store_word(wd, &igs[posn]);
    }
    fclose(fp);
  }
}


void make_lower(char wd[])
/* Convert any uppercase letters in a word to lowercase */
{
  int i = 0;
  while (wd[i] != '\0') {
    if (isupper(wd[i]))
      wd[i] = tolower(wd[i]);
    i++;
  }
}


void store_word(char *wd, Words **ws)
/* Add a word (wd) into the linked list (ws), placing it
   in alphabetical order */
{
  Words *curr, *prev;

  prev = NULL;
  curr = *ws;
  while ((curr->count != CSENT) &&
         (strcmp(wd, curr->word) > 0)) {
    prev = curr;
    curr = curr->next;
  }
  if (strcmp(wd, curr->word) == 0)
    curr->count++;
  else {  /* at end, or curr pointing to bigger word */
    if (prev == NULL) {
      *ws = new_word(wd, 1);
      (*ws)->next = curr;
    }
    else {
      prev->next = new_word(wd, 1);
      prev->next->next = curr;
    }
  }
}


void rec_wds(char *fnm, Words *wds[], Words *igs[])
/* Record each word from the file fnm into wds[], 
   ignoring ones in igs[] */
{
  char wd[WSIZE];
  int i, c, posn;
  FILE *fp;

  if ((fp = fopen(fnm, "r")) == NULL)
    printf("Cannot open %s for reading\n", fnm);
  else {
    printf("Reading: %s\n", fnm);
    while ((c = getc(fp)) != EOF)
      if (isalpha(c)) {
        i = 0;
        wd[i++] = c;
        while (((c = getc(fp)) != EOF) &&
               (isalpha(c)) && (i < WSIZE))
          wd[i++] = c;
        wd[i] = '\0';
        make_lower(wd);
        posn = wd[0] - 'a';
        if (ignore_me(wd, igs[posn]) != 1)
          store_word(wd, &wds[posn]);
      }
    fclose(fp);
  }
}


void dirwalk(char *dir, int sing_lvl, Words *wds[], Words *igs[])
/* Process every entry in a directory */
{
  char name[MAX_PATH];
  struct dirent *dp;
  DIR *dfd;

  if ((dfd = opendir(dir)) == NULL) {
    printf("Cannot open the directory %s\n", dir);
    return;
  }
  while ((dp = readdir(dfd)) != NULL) {
    if ((strcmp(dp->d_name, ".") == 0)
        || (strcmp(dp->d_name, "..") == 0))
      continue;   /* skip self and parent */
    if ((strlen(dir) + strlen(dp->d_name) + 2) >= MAX_PATH)
      printf("<b>Error</b>: Name %s/%s too long<p>\n", dir, dp->d_name);
    else {
      sprintf(name, "%s/%s", dir, dp->d_name);
      search_files(name, sing_lvl, wds, igs);
    }
  }
  closedir(dfd);
}



void search_files(char *name, int sing_lvl, Words *wds[], Words *igs[])
/* Recursively search a subdirectory or process a file */
{
  struct stat stbuf;

  if (stat(name, &stbuf) == -1) {
    printf("Cannot read stat structure for %s\n", name);
    return;
  }
  if (((stbuf.st_mode & S_IFMT) == S_IFDIR) &&
      (sing_lvl == 0))
    dirwalk(name, sing_lvl, wds, igs);
  else if (html_cmp(name) == 0)
    rec_wds(name, wds, igs);
}



int ignore_me(char *wd, Words *ws)
/* Criteria for ignoring a word:
    - length < MINLEN
    - presence in ws list
*/
{
  if (strlen(wd) < MINLEN)
    return 1;

  while (ws->count != CSENT) {
    if (strcmp(wd, ws->word) == 0)
      return 1;
    ws = ws->next;
  }
  return 0;
}


void print_words(int mnum, Words *ws[])
/* Print out the words in ws[] */
{
  int i;

  printf("\nWord Occurrences\n\n");
  if (mnum > 0)
    printf("%d or more occurrences printed\n", mnum);

  for (i=0; i < NUM_WORDS; i++) {
    printf("%c\n", 'a'+ i); 
    printf("===\n");
    print_occurrences(mnum, ws[i]);
  }
}


void print_occurrences(int mnum, Words *ws)
/* Print out each word in the ws list, but ignore ones
   with a count < mnum */
{
  int cols = 0;

  if (ws->count == CSENT)
    printf("-- no entries --\n\n");
  else {
    while (ws->count != CSENT) {
      if (ws->count >= mnum) {
        printf("%10s: %d  ", ws->word, ws->count);
        if (cols == (COL_WIDTH-1)) {
          putchar('\n');
          cols = 0;
        }
        else
          cols++;
      }
      ws = ws->next;
    }
    printf("\n\n");
  }
}

