
/*  Copyright (C) 1995-1996 Andrew Davison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Usage:
      usedby [-s] [-m NUM] html_file | directory w1 [w2] ... [wMAX]
   or
      usedby [-s] [-m NUM] html_file | directory txt_file

   e.g.    usedby -m 3 "." foo goo
           usedby -s humour hello
           usedby "." words.txt

   Date: May 1st, 1996

   Search through an HTML file or a directory. If the
   search is on a directory it will recursively search 
   all the subdirectories unless the -s option is set.

   The search prints out the files containing 1 or more of
   the words given on the command line (up to a max of MAX_MATCH).
   The words may also be loaded from a .txt file specified on
   the command line.

   The recorded words and their occurrences are printed at
   the end in alphabetical order. No distinction is made
   between upper and lower case.
  
   If a minimum NUM has been specified then only occurrences
   >= NUM are printed *per file*.

   Author: Andrew Davison

   Address (May 1996):
   The Department of Computer Science
   The University of Melbourne
   Parkville, Victoria 3052, Australia
   Email: ad@cs.mu.oz.au
   Tel: +61 3 9287 9172
   Fax: +61 3 9348 1184
   URL: http://www.cs.mu.oz.au/~ad
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>       /* flags for read and write */
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>      /* directory features */


#define MAX_MATCH 30     /* maximum number of matches */
#define WSIZE 20         /* the maximum size of a word */
#define MAX_PATH 1024    /* max length of a full filename */


typedef struct word_occ Words;

char *process_args(int argc, char *argv[], int *sl, int *isf, 
                        int *mnum, char *matches[], int *mtno);
int get_matches(int argc, char *argv[], int posn, char *matches[]);
int load_matches(char *fnm, char *matches[]);
void list_matches(char *matches[], int mtno);
int ext_cmp(char *fname, char *ext);
void rec_wds(char *fnm, char *matches[], int mtno, int mnum);
void make_lower(char wd[]);
int matching(char *wd, char *matches[], int mtno);
void print_occurrences(char *fnm, int wd_matches[], int mtno, int mnum);
int any_match(int wd_matches[], int mtno, int mnum);
void dirwalk(char *dir, int sing_lvl, char *matches[], int mtno, int mnum);
void search_files(char *name, int sing_lvl, char *matches[],
                                             int mtno, int mnum);


int main(int argc, char *argv[])
{
  int sing_lvl, is_hfile, mnum, match_no;
  char *srch_nm;
  char *matches[MAX_MATCH];  /* for matching words */

  srch_nm = process_args(argc, argv, &sing_lvl, &is_hfile, 
                               &mnum, matches, &match_no);
  list_matches(matches, match_no);
  if (is_hfile)
    rec_wds(srch_nm, matches, match_no, mnum);
  else
    dirwalk(srch_nm, sing_lvl, matches, match_no, mnum);

  return 1;
}


char *process_args(int argc, char *argv[], int *sl, int *isf, 
                               int *mnum, char *matches[], int *mtno)
/* From the command line, extract the single level flag (sl), 
   the name of the file or directory (setting the flag isf in the process), 
   and the minimum occurrence number (mnum).
   Up to MAX_MATCH matching words come last, the actual number in mtno.
*/
{
  char *name;
  int posn;

  *sl =0;
  *mnum = 0;

  if (argc < 3) {
    printf("Usage: occurs [-s] [-m NUM] html_file | directory w1 [w2]...\n");
    exit(1);
  }

  if (strcmp(argv[1], "-s") == 0) {
    *sl = 1;
    if (strcmp(argv[2], "-m") == 0) {
      *mnum = atoi(argv[3]);
      name = argv[4];
      posn = 5;
    }
    else {
      name = argv[2];
      posn = 3;
    }
  }
  else if (strcmp(argv[1], "-m") == 0) {
    *mnum = atoi(argv[2]);
    name = argv[3];
    posn = 4;
  }
  else {
    name = argv[1];
    posn = 2;
  }
  if (ext_cmp(argv[posn], "txt") == 0)
    *mtno = load_matches(argv[posn], matches);
  else
    *mtno = get_matches(argc, argv, posn, matches);

  if (*mnum > 0)
    printf("%d or more occurrences printed\n", *mnum);

  if (ext_cmp(name, "html") == 0) {
    *isf = 1;
    printf("Processing the HTML file \"%s\"\n", name);
  }
  else {
    *isf = 0;
    printf("Processing the directory \"%s\"", name);
    if (*sl == 1)
      printf(" ONLY\n\n");
    else
      printf(" AND all of its subdirectories\n\n");
  }
  return name;
}


int get_matches(int argc, char *argv[], int posn, char *matches[])
/* Build the matches array by reading words from the command line */
{
  int match_no = 0;

  while ((posn < argc) && (match_no < MAX_MATCH)) {
    matches[match_no] = argv[posn];
    make_lower(matches[match_no]);
    match_no++; posn++;
  }

  if (match_no == 0) {
    printf("Error: you must supply at least 1 matching word\n");
    printf("Usage: occurs [-s] [-m NUM] html_file | directory w1 [w2]...\n");
    exit(1);
  }
  if ((match_no == MAX_MATCH) && (posn < argc)) {
    printf("Error: %d matching word(s) ignored; %d accepted\n",
                               (argc - posn), MAX_MATCH);
  }
  return match_no;
}


int load_matches(char *fnm, char *matches[])
/* Build the matches array by reading words from a file */
{
  char wd[WSIZE];
  int match_no = 0;
  FILE *fp;

  if ((fp = fopen(fnm, "r")) == NULL) {
    printf("Cannot read matching words file: %s\n", fnm);
    exit(1);
  }
  else {
    printf("Reading matching words file: %s\n", fnm);
    while ((fscanf(fp, "%s", wd) != EOF) &&
           (match_no < MAX_MATCH)) {
      make_lower(wd);
      matches[match_no] = (char *) malloc(sizeof(char)*(strlen(wd)+1));
      strcpy(matches[match_no], wd);
      match_no++;
    }
    fclose(fp);
    if (match_no == 0) {
      printf("Error: you must supply at least 1 matching word\n");
      exit(1);
    }
    if (match_no ==  MAX_MATCH)
      printf("Error: maximum matching words is %d; others ignored\n",
                                     MAX_MATCH);
  }
  return match_no;
}


void list_matches(char *matches[], int mtno)
{
  int i;

  printf("Matching against %d word(s):    ", mtno);
  for (i=0; i < mtno; i++)
    printf("%8s ", matches[i]);
  printf("\n\n");
}


int ext_cmp(char *fname, char *ext)
/* Compare the file name end with extension in ext*/
{
  int i=0;

  if (fname[0] == '.')
    i = 1;       /* to skip possible "." at beginning */
  while ((fname[i] != '\0') && (fname[i] != '.'))
    i++;
  if (fname[i] == '.')
    return strcmp(&fname[i+1], ext);
  else
    return 1;
}



void rec_wds(char *fnm, char *matches[], int mtno, int mnum)
/* For each file fnm, match its words against the ones in matches[]
   and print out the number of matches. */
{
  int wd_matches[MAX_MATCH] = {0};
  char wd[WSIZE];
  int i, c, posn;
  FILE *fp;

  if ((fp = fopen(fnm, "r")) == NULL)
    printf("Cannot open %s for reading\n", fnm);
  else {
    /* printf("Reading %s: ", fnm); */
    while ((c = getc(fp)) != EOF)
      if (isalpha(c)) {
        i = 0;
        wd[i++] = c;
        while (((c = getc(fp)) != EOF) &&
               (isalpha(c)) && (i < WSIZE))
          wd[i++] = c;
        wd[i] = '\0';
        make_lower(wd);
        if ((posn = matching(wd, matches, mtno)) != -1)
          wd_matches[posn]++;
      }
    fclose(fp);
    print_occurrences(fnm, wd_matches, mtno, mnum);
  }
}


void make_lower(char wd[])
/* Convert any uppercase letters in a word to lowercase */
{
  int i = 0;
  while (wd[i] != '\0') {
    if (isupper(wd[i]))
      wd[i] = tolower(wd[i]);
    i++;
  }
}


int matching(char *wd, char *matches[], int mtno)
/* Return position of wd in matches[], or -1 */
{
  int i;
  for (i=0; i < mtno; i++)
    if (strcmp(wd, matches[i]) == 0)
      return i;
  return -1;
}


void print_occurrences(char *fnm, int wd_matches[], int mtno, int mnum)
/* Print out the file and matches. 
   Ignore it if all its matches are below mnum.
*/
{
  int i;

  if (any_match(wd_matches, mtno, mnum) == 1) {
    printf("%-30s  ", fnm);
    for(i=0; i < mtno; i++)
      printf("%8d", wd_matches[i]);
    putchar('\n');
  }
}


int any_match(int wd_matches[], int mtno, int mnum)
{
  int i;
  for(i=0; i < mtno; i++)
    if ((wd_matches[i] >= mnum) && (wd_matches[i] > 0))
      return 1;
  return 0;
}



void dirwalk(char *dir, int sing_lvl, char *matches[], int mtno, int mnum)
/* Process every entry in a directory */
{
  char name[MAX_PATH];
  struct dirent *dp;
  DIR *dfd;

  if ((dfd = opendir(dir)) == NULL) {
    printf("Cannot open the directory %s\n", dir);
    return;
  }
  while ((dp = readdir(dfd)) != NULL) {
    if ((strcmp(dp->d_name, ".") == 0)
        || (strcmp(dp->d_name, "..") == 0))
      continue;   /* skip self and parent */
    if ((strlen(dir) + strlen(dp->d_name) + 2) >= MAX_PATH)
      printf("<b>Error</b>: Name %s/%s too long<p>\n", dir, dp->d_name);
    else {
      sprintf(name, "%s/%s", dir, dp->d_name);
      search_files(name, sing_lvl, matches, mtno, mnum);
    }
  }
  closedir(dfd);
}



void search_files(char *name, int sing_lvl, char *matches[], 
                                             int mtno, int mnum)
/* Recursively search a subdirectory or process a file */
{
  struct stat stbuf;

  if (stat(name, &stbuf) == -1) {
    printf("Cannot read stat structure for %s\n", name);
    return;
  }
  if (((stbuf.st_mode & S_IFMT) == S_IFDIR) &&
      (sing_lvl == 0))
    dirwalk(name, sing_lvl, matches, mtno, mnum);
  else if (ext_cmp(name, "html") == 0)
    rec_wds(name, matches, mtno, mnum);
}


